unit Cltn;

interface

(*
  Collections.
  These interfaces define a general-purpose set of collections.

  Each collection interface might have one or more different classes
  that implement the interface, giving different performance characteristics,
  e.g., THashSet or TTreeSet.

  In order to appear in a collection, an object must implement the
  ICollectible interface. The CompareTo methods returns an integer result:
    < 0 means Self < Obj
    = 0 means the objects are identical
    > 0 means Self > Obj

  A hash value can be any non-negative number. A sample hash function
  for strings is given as the HashString function. For two given objects,
  if the IsEqual function returns True, then their HashValue functions
  must return the same value. The reverse is not necessarily true:
  two objects with the same hash value are not necessarily equal.

  You can have multiple enumerators active on a given collection, but
  you should not modify the collection while the enumerator is active.
  If you must modify the collection, reset the enumerator.

  The enumerator interfaces follow Microsoft's COM convention for
  the Next, Skip, Reset, and Clone methods, but also implement
  some additional methods to make them more useful in Delphi programs.

  When implementing any of these collection interfaces, use the the
  collection exceptions defined in this unit.

  Copyright  1999 Tempest Software, Inc.
*)

uses Windows, SysUtils, Classes;

type
  THashValue = 0..MaxInt;
  ICollectible = interface
  ['{618977E1-1ADA-11D3-B1A8-00105AA9C2AD}']
    // Return a value < 0 if Self<Obj
    //                = 0 if Self=Obj
    //                > 0 if Self>Obj
    function CompareTo(const Obj: ICollectible): Integer;
    // Return a unique hash value. If two objects are equal (CompareTo=0)
    // their hash values must also be equal, but the reverse is not
    // necessarily true.
    function HashValue: THashValue;
  end;

  // For use in Maps. A Map is a Set where every member of the set
  // is an IAssociation. An assocation associates a key and a value.
  // The Map uses the key to determine set membership; the value just
  // tags along for good luck.
  IAssociation = interface
  ['{6CF8D12F-16D3-11D3-B1A8-00105AA9C2AD}']
    procedure GetKey(out Key: ICollectible);
    procedure GetValue(out Value: IUnknown);
    procedure SetValue(Value: IUnknown);
  end;

  // The array types are used by enumerators to fetch multiple items from
  // a collection. Usually, you will use the dynamic arrays, but you can
  // use the static arrays, if you prefer.
  TCollectibleArray = array of ICollectible;
  TAssociationArray = array of IAssociation;

  TStaticCollectibleArray = array[0..MaxListSize-1] of ICollectible;
  TStaticAssociationArray = array[0..MaxListSize-1] of IAssociation;

  // User-supplied function; return True to continue, False to abort.
  TCollectionProc = function (Obj: ICollectible): Boolean of object;
  TAssociationProc = function (Key: ICollectible; Value: IUnknown): Boolean of object;

  IEnumerator = interface;
  ICollection = interface
  ['{618977E3-1ADA-11D3-B1A8-00105AA9C2AD}']
    // Add one or more items to the collection.
    procedure Add(Obj: ICollectible);
    procedure AddAll(Collection: ICollection);

    // Remove all items from the collection.
    procedure Clear;

    // Make a copy of this collection. Do not copy the items in the collection,
    // but keep references to the same items.
    procedure Clone(out Collection: ICollection);

    // Return true when the collection contains the object.
    function Contains(const Obj: ICollectible): Boolean;
    // Return true if this collection contains all of the items in "Collection",
    // in any order.
    function ContainsAll(const Collection: ICollection): Boolean;

    // Count the number of items matching the given item.
    function CountMatching(const Obj: ICollectible): Integer;

    // Return a new enumerator for this collection. Do not modify the collection
    // while the enumerator is active. You can have any number of enumerators
    // for each collection.
    procedure Enumerator(out Enum: IEnumerator); overload;

    // Look for the first matching object and get a reference to the
    // object in the collection. Return True if found, False if not found.
    // Set Found to nil if not found.
    function Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean;

    // Call Proc for each item in the collection.
    procedure ForEach(Proc: TCollectionProc); overload;

    // Return the number of items in the collection.
    function GetCount: Integer;

    // Tell the collection how to handle duplicate items. Some collections
    // do not let you change the duplicates behavior, and will raise an
    // exception if you try to set Duplicates to an invalid value.
    function GetDuplicates: TDuplicates;
    procedure SetDuplicates(Dup: TDuplicates);

    // Return True if Count=0, False if Count>0.
    function IsEmpty: Boolean;

    // Return true if the other collection has the same contents.
    function IsEqual(const Collection: ICollection): Boolean;

    // Query the collection class and return True if:
    // IsMap: the collection is a map collection
    // IsSorted: the collection puts items into sorted order
    // IsOrdered: the collection preserves the order in which items are added
    function IsMap: Boolean;
    function IsSorted: Boolean;
    function IsOrdered: Boolean;

    // Remove the first occurence of an item from the collection.
    // Raise EItemNotFound if it is not in the collection.
    procedure Remove(const Obj: ICollectible);

    // Remove all occurences of an item from the collection.
    // Do nothing if the item is not in the collection.
    procedure RemoveAll(const Obj: ICollectible); overload;

    // Remove all occurences of all the items in Collection.
    // Raise an exception if any item is not in this collection.
    procedure RemoveAll(const Collection: ICollection); overload;

    // Remove all objects EXCEPT those in "Collection".
    procedure RetainAll(const Collection: ICollection);

    // Return a dynamic array that contains the same items as this collection.
    function ToArray: TCollectibleArray;

    property Count: Integer read GetCount;
    property Duplicates: TDuplicates read GetDuplicates write SetDuplicates;
  end;

  IEnumerator = interface
  ['{618977E4-1ADA-11D3-B1A8-00105AA9C2AD}']
    // Make a copy of this iterator, iterating over the same collection,
    // from the same position.
    procedure Clone(out Enum: IEnumerator);

    // Return True if another item is available to be retrieved by Next,
    // or False if the next call to Next will fail.
    function HasNext: Boolean;

    // Get the next item or items in the collection. Choose the dynamic array
    // or static array for returning multiple items. NumFetched tells you how
    // many items were fetched. For a dynamic array, the array size tells you
    // how many items were fetched.
    // The last forms of Next fetches at most one item, returning False
    // if no more items can be fetched.
    procedure Next(Count: Integer; out ResultArray: TCollectibleArray); overload;
    procedure Next(Count: Integer; var ResultArray: TStaticCollectibleArray; var NumFetched: Integer); overload;
    function Next(out Obj: ICollectible): Boolean; overload;

    // Remove the item that was most recently fetched by Next.
    // A subsequent call to Next fetches the item immediately
    // after the item that was removed. If Next failed or was not
    // called first, Remove raises an exception.
    procedure Remove;

    // Reset the enumerator so Next will fetch the first item.
    procedure Reset;

    // Try to skip over Count items in the collection. Return the number
    // of items actually skipped.
    function Skip(Count: Integer = 1): Integer;
  end;

  // A set is the most basic real collection
  // Some sets force Duplicates = dupIgnore or dupError, but others
  // implement "multiset" functionality and let you set Duplicates
  // to dupAccept.
  ISet = interface(ICollection)
  ['{6CF8D136-16D3-11D3-B1A8-00105AA9C2AD}']
    // Intersect with Collection, i.e., delete all items that
    // are NOT in both collections.
    procedure Intersect(const Collection: ICollection);
    // Union with Collection, i.e., add all items from Collection
    // that are not already present in this collection.
    procedure Union(const Collection: ICollection);
    // Subtract from this collection all items that are in Collection.
    procedure Subtract(const Collection: ICollection);
  end;
  IEnumSet = interface(IEnumerator)
  ['{6CF8D137-16D3-11D3-B1A8-00105AA9C2AD}']
  end;

  // Sometimes, you need to sort a list differently from the default
  // ICollectible.CompareTo function. In that case, you can use
  // an explicit comparison function.
  TListCompareFunc = function(const A, B: ICollectible): Integer of object;

  // A list contains items in a specific order.
  IList = interface(ICollection)
  ['{618977E5-1ADA-11D3-B1A8-00105AA9C2AD}']
    function GetItems(Index: Integer): ICollectible;
    procedure SetItems(Index: Integer; Item: ICollectible);
    function IndexOf(const Item: ICollectible): Integer;

    // A list is special and allows adding and deleting an item
    // at a specific index.
    procedure Delete(Index: Integer);
    procedure Insert(Index: Integer; Obj: ICollectible);
    procedure GetItem(Index: Integer; out Item: ICollectible);

    // Exchange the list items at indices I and J.
    procedure Exchange(I, J: Integer);
    // Move the item at CurIndex to NewIndex.
    procedure Move(CurIndex, NewIndex: Integer);

    procedure Sort; overload;
    procedure Sort(Compare: TListCompareFunc); overload;

    property Items[Index: Integer]: ICollectible read GetItems write SetItems; default;
  end;
  IEnumList = interface(IEnumerator)
  ['{618977E6-1ADA-11D3-B1A8-00105AA9C2AD}']
  end;

  // The TCltnSort class has some methods that help other
  // classes implement the IList.Sort methods.
  TCltnSort = class
  protected
    class procedure QuickSort(const List: IList; Lo, Hi: Integer; Compare: TListCompareFunc);
  public
    class function DefaultCompare(const A, B: ICollectible): Integer;
    class procedure Sort(const List: IList; Compare: TListCompareFunc = nil);
  end;

  // A stack also has a specific order, but you access it from one end
  // only: the top.
  IStack = interface(ICollection)
  ['{E7A78232-1C36-11D3-B1A8-00105AA9C2AD}']
    // Get the top of stack. Return True for success, False if stack if empty.
    function GetTop(out Top: ICollectible): Boolean;
    // Add an item to the stack.
    procedure Push(Obj: ICollectible);
    // Get the top of stack and remove it.
    function Pop(out Top: ICollectible): Boolean;
  end;

  // A map stores key/value pairs. Lookup, remove, etc., by key or value.
  // Access by key is fast; access by value is possible, but requires
  // a linear search over the entire collection. The value can usually be
  // any IUnknown, but if you want to search by value, the value must
  // implement ICollectible.
  // The inherited method--Add, Contains, Find, and Remove--take associations
  // as arguments or return IAssociation objects as results.
  IEnumMap = interface;
  IMap = interface(ICollection)
  ['{6CF8D130-16D3-11D3-B1A8-00105AA9C2AD}']
    procedure Add(Assoc: IAssociation); overload;
    procedure Add(Key: ICollectible; Value: IUnknown); overload;
    function ContainsKey(const Key: ICollectible): Boolean;
    function ContainsValue(const Value: ICollectible): Boolean;
    procedure Enumerator(out Enum: IEnumMap); overload;
    function FindKey(const Key: ICollectible; out Assoc: IAssociation): Boolean;
    function FindValue(const Value: ICollectible; out Assoc: IAssociation): Boolean;
    procedure RemoveKey(const Key: ICollectible);
    procedure RemoveValue(const Value: ICollectible);

    // Call Proc for each association in the map.
    procedure ForEach(Proc: TAssociationProc); overload;

    // Get the value for a key, or raise an EItemNotFound exception.
    function GetPropValue(const Key: ICollectible): IUnknown;
    procedure SetPropValue(const Key: ICollectible; Value: IUnknown);

    // Get a value for a key, and return True if found or False if not found.
    function GetValue(const Key: ICollectible; out Value: IUnknown): Boolean;

    // Get a set of just the keys.
    procedure GetKeys(out Keys: ISet);
    // Get a collection of just the values. The values must implement
    // ICollectible.
    procedure GetValues(out Values: ICollection);

    property Value[const Key: ICollectible]: IUnknown read GetPropValue write SetPropValue; default;
  end;

  // The inherited Next methods fetch associations, but cast as ICollectible.
  // For your convenience, you can use the additional Next methods, which
  // explicitly use IAssociation results. The NextKey methods fetch the
  // next association, but returns only the key part.
  IEnumMap = interface(IEnumerator)
  ['{6CF8D131-16D3-11D3-B1A8-00105AA9C2AD}']
    procedure  Next(Count: Integer; out ResultArray: TAssociationArray); overload;
    procedure Next(Count: Integer; var ResultArray: TStaticAssociationArray; var NumFetched: Integer); overload;
    function Next(out Obj: IAssociation): Boolean; overload;
    function NextKey(out Obj: ICollectible): Boolean; overload;
  end;


  // As a convenience for implementing collection classes, use TCollection
  // as an abstract base class. It implements several of the methods in a
  // generic way.
  TCollectionClass = class of TCollection;
  TCollection = class(TInterfacedObject, ICollection)
  protected
    // The following method has an inefficient implementation, so
    // derived classes should override it, if possible.
    function GetCount: Integer; virtual;

    // The default is for collections to accept duplicate members.
    // Derived classes should override the following methods if they
    // give the programmer control or do something different.
    function GetDuplicates: TDuplicates; virtual;
    procedure SetDuplicates(Dup: TDuplicates); virtual;


    procedure DuplicateError(Item: ICollectible);
    procedure ItemNotFound(Item: ICollectible);
    procedure ValueNotFound(Item: ICollectible);
    procedure IndexOutOfRange(Index: Integer);
    procedure NotImplemented(const MethodName: string);

  public
    constructor Create; virtual;

    // Derived classes must override the following four methods:
    procedure Add(Obj: ICollectible); overload; virtual; abstract;
    procedure Enumerator(out Enum: IEnumerator); overload; virtual; abstract;
    function Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean; virtual; abstract;
    procedure Remove(const Obj: ICollectible); virtual; abstract;

    // The following methods have inefficient implementations, so derived
    // classes should override them, if possible.
    procedure Clear; virtual;
    function CountMatching(const Obj: ICollectible): Integer; virtual;
    procedure RetainAll(const Collection: ICollection); virtual;

    // The default implementation of the following methods always returns False.
    // Override any or all of them to return True when appropriate.
    function IsMap: Boolean; virtual;
    function IsSorted: Boolean; virtual;
    function IsOrdered: Boolean; virtual;

    // The following have convenient implementations. Most derived classes
    // will not need to override these methods, but can do so if the derived
    // class has a better way of doing the same thing.
    procedure AddAll(Collection: ICollection); virtual;
    procedure Clone(out Collection: ICollection); virtual;
    function Contains(const Obj: ICollectible): Boolean; virtual;
    function ContainsAll(const Collection: ICollection): Boolean; virtual;
    procedure First(out Item: ICollectible); virtual;
    procedure ForEach(Proc: TCollectionProc); overload; virtual;
    function IsEmpty: Boolean; virtual;
    function IsEqual(const Collection: ICollection): Boolean; virtual;
    procedure RemoveAll(const Obj: ICollectible); overload; virtual;
    procedure RemoveAll(const Collection: ICollection); overload; virtual;
    function ToArray: TCollectibleArray; virtual;

    property Count: Integer read GetCount;
    property Duplicates: TDuplicates read GetDuplicates write SetDuplicates;
  end;

  // As a convenience for creating new collection implementations,
  // TFilterCollection implements the ICollection protocol by passing
  // all the methods to a different collection. See TMap and TSet for
  // ways of using TFilterCollection.
  TFilterCollection = class(TCollection)
  private
    fCollection: ICollection;
  protected
    // Derived classes must override the following method to
    // create the collection that stores the filtered collection.
    procedure CreateCollection(out Collection: ICollection); virtual; abstract;

    function GetCount: Integer; override;
    function GetDuplicates: TDuplicates; override;
    procedure SetDuplicates(Dup: TDuplicates); override;

    property Collection: ICollection read fCollection;
  public
    procedure Add(Obj: ICollectible); overload; override;
    procedure Clear; override;
    function Contains(const Obj: ICollectible): Boolean; override;
    procedure Enumerator(out Enum: IEnumerator); overload; override;
    function Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean; override;
    function IsEmpty: Boolean; override;
    function IsSorted: Boolean; override;
    function IsOrdered: Boolean; override;
    procedure Remove(const Obj: ICollectible); override;
    procedure RemoveAll(const Obj: ICollectible); overload; override;
    procedure RemoveAll(const Collection: ICollection); overload; override;
    procedure RetainAll(const Collection: ICollection); override;
    function ToArray: TCollectibleArray; override;
  end;

  // Any set can be made into a map by storing associations as the set members.
  // This class can take any kind of collection and turn it into a map.
  // To use TMap, derive an implementation-specific class and define
  // a public constructor that calls the protected constructor, passing
  // a collection instance as an argument.
  TMap = class(TFilterCollection, IMap)
  protected
    // Derived classes must override the following methods to
    // create the collections that store the map and are returned
    // from GetKeys and GetValues. TMap takes care of filling the collections.
    procedure CreateCollection(out Collection: ICollection); override; abstract;
    procedure CreateValuesCollection(out Collection: ICollection); virtual; abstract;
    procedure CreateKeysCollection(out Collection: ISet); virtual; abstract;

    function GetPropValue(const Key: ICollectible): IUnknown; virtual;
    procedure SetPropValue(const Key: ICollectible; Value: IUnknown); virtual;

    property Collection: ICollection read fCollection;
  public
    constructor Create; override;

    // Derived classes might want to provide a custom enumerator
    // instead of TEnumMap.
    procedure Enumerator(out Enum: IEnumMap); overload; virtual;
    procedure Enumerator(out Enum: IEnumerator); overload; override;

    procedure Add(Assoc: IAssociation); overload; virtual;
    procedure Add(Key: ICollectible; Value: IUnknown); overload; virtual;
    procedure Add(Obj: ICollectible); overload; override;
    function ContainsKey(const Key: ICollectible): Boolean; virtual;
    function ContainsValue(const Value: ICollectible): Boolean; virtual;
    function FindKey(const Key: ICollectible; out Assoc: IAssociation): Boolean; virtual;
    function FindValue(const Value: ICollectible; out Assoc: IAssociation): Boolean; virtual;
    procedure ForEach(Proc: TAssociationProc); overload; virtual;
    procedure GetKeys(out Keys: ISet); virtual;
    function GetValue(const Key: ICollectible; out Value: IUnknown): Boolean; virtual;
    procedure GetValues(out Values: ICollection); virtual;
    function IsMap: Boolean; override;
    procedure RemoveKey(const Key: ICollectible); virtual;
    procedure RemoveValue(const Value: ICollectible); virtual;

    property Value[const Key: ICollectible]: IUnknown read GetPropValue write SetPropValue; default;
  end;

  // Derive a class from TSet to implement a specific kind of set.
  TSet = class(TFilterCollection, ISet)
  protected
    procedure CreateCollection(out Collection: ICollection); override; abstract;
    // TSet guarantees unique keys. A derived class can override GetDuplicates
    // and SetDuplicates to implement a multiset or bag.
    function GetDuplicates: TDuplicates; override;
    procedure SetDuplicates(Dup: TDuplicates); override;
  public
    constructor Create; override;

    procedure Add(Obj: ICollectible); overload; override;
    procedure Enumerator(out Enum: IEnumSet); overload; virtual;
    procedure Intersect(const Collection: ICollection); virtual;
    procedure Union(const Collection: ICollection); virtual;
    procedure Subtract(const Collection: ICollection); virtual;
  end;

  // As a convenience for implementing enumerators, use TEnumerator as an
  // abstract base class.
  TEnumerator = class(TInterfacedObject, IEnumerator)
  private
    fCollection: TCollection;
  protected
    // Derived classes must override the following methods.
    function HasNext: Boolean; virtual; abstract;
    procedure Remove; virtual; abstract;
    procedure Reset; virtual; abstract;
    function Next(out Obj: ICollectible): Boolean; overload; virtual; abstract;

    // Derived classes should override the following method, and be sure
    // to call the inherited Clone method.
    procedure Clone(out Enum: IEnumerator); overload; virtual;

    constructor Create(Collection: TCollection); virtual;

    procedure CantRemoveError;
    procedure GetCollection(out Collection: ICollection);
    property CollectionObject: TCollection read fCollection write fCollection;
  public
    destructor Destroy; override;
    procedure Next(Count: Integer; out ResultArray: TCollectibleArray); overload; virtual;
    procedure Next(Count: Integer; var ResultArray: TStaticCollectibleArray; var NumFetched: Integer); overload; virtual;
    function Skip(Count: Integer = 1): Integer; virtual;
  end;

  TEnumMap = class(TEnumerator, IEnumMap)
  private
    fEnumerator: IEnumerator;
  protected
    property Enumerator: IEnumerator read fEnumerator;
  public
    constructor Create(Collection: TCollection); override;

    procedure Clone(out Enum: IEnumerator); overload; override;
    function HasNext: Boolean; override;
    function Next(out Item: ICollectible): Boolean; overload; override;
    procedure Remove; override;
    procedure Reset; override;

    procedure Next(Count: Integer; out ResultArray: TAssociationArray); overload;
    procedure Next(Count: Integer; var ResultArray: TStaticAssociationArray; var NumFetched: Integer); overload;
    function Next(out Obj: IAssociation): Boolean; overload;
    function NextKey(out Key: ICollectible): Boolean;
  end;

  // The following interfaces and classes provide common implementations
  // for strings, objects, and numbers.
  IString = interface(ICollectible)
  ['{618977E9-1ADA-11D3-B1A8-00105AA9C2AD}']
    function GetValue: string;
    procedure SetValue(const Value: string);

    property Value: string read GetValue write SetValue;
  end;

  TString = class(TInterfacedObject, ICollectible, IString)
  private
    fValue: string;
    function GetValue: string;
    procedure SetValue(const Value: string);
  public
    constructor Create(Value: string = ''); overload;
    constructor Create(Value: TString); overload;
    destructor Destroy; override;
    function CompareTo(const Obj: ICollectible): Integer;
    function HashValue: THashValue;

    property Value: string read fValue write fValue;
  end;

  IObject = interface(ICollectible)
  ['{618977EA-1ADA-11D3-B1A8-00105AA9C2AD}']
    function GetValue: TObject;
    procedure SetValue(Value: TObject);
    property Value: TObject read GetValue write SetValue;
  end;

  TWrapper = class(TInterfacedObject, ICollectible, IObject)
  private
    fObject: TObject;
  public
    constructor Create(Obj: TObject); overload;
    constructor Create(Wrapper: TWrapper); overload;
    function GetValue: TObject;
    procedure SetValue(Value: TObject);

    function CompareTo(const Obj: ICollectible): Integer;
    function HashValue: THashValue;

    property Value: TObject read fObject write fObject;
  end;

  IInterface = interface
  ['{1336B5A1-4D2D-11D3-B1B2-00105AA9C2AD}']
    function GetInterface: IUnknown;
    procedure SetInterface(const Value: IUnknown);
    property Intf: IUnknown read GetInterface write SetInterface;
  end;

  TInterfaceWrapper = class(TInterfacedObject, ICollectible, IInterface)
  private
    fInterface: IUnknown;
  protected
    function GetInterface: IUnknown;
    procedure SetInterface(const Value: IUnknown);
  public
    constructor Create(Intf: IUnknown);

    function CompareTo(const Obj: ICollectible): Integer;
    function HashValue: THashValue;

    property Intf: IUnknown read fInterface write SetInterface;
  end;

  IInteger = interface(ICollectible)
  ['{D77E51D1-1EB3-11D3-B1A8-00105AA9C2AD}']
    function GetValue: Integer;
    procedure SetValue(Value: Integer);
    property Value: Integer read GetValue write SetValue;
  end;

  TInteger = class(TInterfacedObject, ICollectible, IInteger)
  private
    fValue: Integer;
    function GetValue: Integer;
    procedure SetValue(Value: Integer);
  public
    constructor Create(Value: Integer); overload;

    function CompareTo(const Obj: ICollectible): Integer;
    function HashValue: THashValue;

    property Value: Integer read fValue write fValue;
  end;

  IDouble = interface(ICollectible)
  ['{D77E51D1-1EB3-11D3-B1A8-00105AA9C2AD}']
    function GetValue: Double;
    procedure SetValue(Value: Double);
    property Value: Double read GetValue write SetValue;
  end;

  TDouble = class(TInterfacedObject, ICollectible, IDouble)
  private
    fValue: Double;
    function GetValue: Double;
    procedure SetValue(Value: Double);
  public
    constructor Create(Value: Double); overload;

    function CompareTo(const Obj: ICollectible): Integer;
    function HashValue: THashValue;

    property Value: Double read fValue write fValue;
  end;

  // To sort in reverse order, wrap all the objects in TReverseWrappers.
  // This class simply reverses the sign that CompareTo returns.
  TReverseWrapper = class(TInterfacedObject, ICollectible)
  private
    fItem: ICollectible;
  public
    constructor Create(Item: ICollectible);
    function CompareTo(const Obj: ICollectible): Integer;
    procedure GetItem(out Item: ICollectible);
    function HashValue: THashValue;
    property Item: ICollectible read fItem;
  end;

  TAssociation = class(TInterfacedObject, IAssociation, ICollectible)
  private
    fKey: ICollectible;
    fValue: IUnknown;
  public
    constructor Create(Key: ICollectible; Value: IUnknown);
    function HashValue: THashValue;
    function CompareTo(const Obj: ICollectible): Integer;
    procedure GetKey(out Key: ICollectible);
    procedure GetValue(out Value: IUnknown);
    procedure SetValue(Value: IUnknown);

    property Key: ICollectible read fKey;
    property Value: IUnknown read fValue write fValue;
  end;

  // Store a string and an object, for use in TCltnStrings.
  TStringItem = class(TString, IObject)
  private
    fObject: TObject;
    function GetObject: TObject;
    procedure SetObject(Value: TObject);
    function IObject.GetValue = GetObject;
    procedure IObject.SetValue = SetObject;
  public
    constructor Create(const S: string; Obj: TObject = nil);
  end;

  // TCltnStrings follows the TStrings protocol, so you can use it
  // anywhere in Delphi that calls for TStrings. It can use any IList
  // collection to store the strings. By default, each item in the
  // collection is a TStringItem object, but a derived class can
  // override CreateItem to change this.
  // The default collection is TArrayList, but a derived class
  // can override CreateCollection to use a different collection.
  TCltnStrings = class(TStrings)
  private
    fCollection: IList;
  protected
    procedure CreateCollection(out Collection: IList); virtual;
    procedure CreateItem(out Item: ICollectible; const S: string); virtual;

    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure PutObject(Index: Integer; AObject: TObject); override;

    property Collection: IList read fCollection;
  public
    constructor Create;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    function IndexOf(const S: string): Integer; override;
    procedure Insert(Index: Integer; const S: string); override;
    procedure Sort; overload;
    procedure Sort(Compare: TListCompareFunc); overload;
  end;
  TSortedCltnStrings = class(TCltnStrings)
  protected
    procedure CreateCollection(out Collection: IList); override;
  end;

function MakeAssociation(Key: ICollectible; Value: IUnknown = nil): IAssociation;
function MakeString(const Str: string): IString;
function MakeWrapper(Obj: TObject): IObject;
function MakeReverse(Obj: ICollectible): ICollectible;
function MakeInteger(Value: Integer): IInteger;
function MakeDouble(Value: Double): IDouble;

function HashString(const Str: string): THashValue;

type
  ECollectionError = class(Exception)
  end;
  ECollectionItemError = class(ECollectionError)
  private
    fItem: ICollectible;
  public
    constructor Create(const Msg: string; Item: ICollectible); overload;
    constructor Create(const Fmt: string; Args: array of const; Item: ICollectible); overload;
    procedure GetItem(out Item: ICollectible);
    property Item: ICollectible read fItem;
  end;

implementation

uses CltnArrays;

resourcestring
  sCollectionChanged = 'Internal error: ToArray failed because collection changed while enumerator was active';
  sDuplicate = '%s: duplicate item';
  sItemNotFound = '%s: item not found';
  sValueNotFound = '%s: value not found';
  sIndexOutOfRange = '%s: index (%d) is out of range';
  sCantRemove = '%s: cannot call Remove except after successful call to Next';
  sNotImplemented = '%s: Method %s is not implemented in this class';
  sFirstOfEmptyCollection = '%s: cannot call First because collection is empty';


function HashString(const Str: string): THashValue;
var
  I: Integer;
begin
  Result := 0;
  for I := 1 to Length(Str) do
{$Q-}
    Result := (Result * 2) xor Ord(Str[I]);
{$Q+}
end;

function MakeAssociation(Key: ICollectible; Value: IUnknown): IAssociation;
begin
  Result := TAssociation.Create(Key, Value);
end;

function MakeString(const Str: string): IString;
begin
  Result := TString.Create(Str)
end;

function MakeWrapper(Obj: TObject): IObject;
begin
  Result := TWrapper.Create(Obj)
end;

function MakeReverse(Obj: ICollectible): ICollectible;
begin
  Result := TReverseWrapper.Create(Obj)
end;

function MakeInteger(Value: Integer): IInteger;
begin
  Result := TInteger.Create(Value)
end;

function MakeDouble(Value: Double): IDouble;
begin
  Result := TDouble.Create(Value)
end;

{ TCollection }

// Nothing special for the constructor to do, but it serves as a
// placeholder for the virtual constructors for concrete collection classes.
constructor TCollection.Create;
begin
  inherited;
end;

// Add all items from Collection to Self.
procedure TCollection.AddAll(Collection: ICollection);
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Collection.Enumerator(Enum);
  while Enum.Next(Item) do
    Add(Item);
end;

// Delete all the items from the collection. You should usually override this
// method to use a more efficient implementation.
procedure TCollection.Clear;
var
  Item: ICollectible;
  Enum: IEnumerator;
begin
  Enumerator(Enum);
  while Enum.Next(Item) do
    Enum.Remove;
end;

// Make a copy of this collection.
procedure TCollection.Clone(out Collection: ICollection);
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Collection := TCollection(ClassType).Create;
  Enumerator(Enum);
  while Enum.Next(Item) do
    Collection.Add(Item);
end;

// Return True if the collection contains Obj.
function TCollection.Contains(const Obj: ICollectible): Boolean;
var
  Item: ICollectible;
begin
  Result := Find(Obj, Item);
end;

// Return True if the collection contains all of the items in Collection.
function TCollection.ContainsAll(const Collection: ICollection): Boolean;
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Result := False;
  Collection.Enumerator(Enum);
  while Enum.Next(Item) do
    if not Contains(Item) then
      Exit;

  Result := True;
end;

// Count the number of items matching Obj.
function TCollection.CountMatching(const Obj: ICollectible): Integer;
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Result := 0;
  Enumerator(Enum);
  while Enum.Next(Item) do
    if Obj.CompareTo(Item) = 0 then
      Inc(Result);
end;

// Return the first member of the collection.
procedure TCollection.First(out Item: ICollectible);
var
  Enum: IEnumerator;
begin
  Enumerator(Enum);
  if not Enum.Next(Item) then
    raise ECollectionError.CreateFmt(sFirstOfEmptyCollection, [ClassName]);
end;

// Call Proc for each item in the collection. If the user-supplied procedure
// returns False, stop enumerating.
procedure TCollection.ForEach(Proc: TCollectionProc);
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Enumerator(Enum);
  while Enum.Next(Item) do
    if not Proc(Item) then
      Exit;
end;

// Stupid, slow implementation of GetCount. You certainly
// want to do better in any concrete collection class.
function TCollection.GetCount: Integer;
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Result := 0;
  Enumerator(Enum);
  while Enum.Next(Item) do
    Inc(Result);
end;

// Default is to accept duplicates, but some derived classes work differently.
function TCollection.GetDuplicates: TDuplicates;
begin
  Result := dupAccept;
end;

procedure TCollection.SetDuplicates(Dup: TDuplicates);
begin
  NotImplemented('SetDuplicates');
end;

// Return True if the collection has no items.
function TCollection.IsEmpty: Boolean;
begin
  Result := Count = 0;
end;

// Return True if the two collections have the same contents.
function TCollection.IsEqual(const Collection: ICollection): Boolean;
begin
  // This isn't always the most efficient implementation,
  // but it's quick and easy to write.
  Result := (Count = Collection.Count) and ContainsAll(Collection);
end;

function TCollection.IsMap: Boolean;
begin
  Result := False
end;

function TCollection.IsOrdered: Boolean;
begin
  Result := False
end;

function TCollection.IsSorted: Boolean;
begin
  Result := False
end;

// Remove all matching items from Self.
procedure TCollection.RemoveAll(const Obj: ICollectible);
begin
  while Contains(Obj) do
    Remove(Obj);
end;

// Remove all the items that are in Collection from Self.
procedure TCollection.RemoveAll(const Collection: ICollection);
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Collection.Enumerator(Enum);
  while Enum.Next(Item) do
    Remove(Item);
end;

// Remove all the items EXCEPT those in Collection.
// Derived classes should find a more efficient implementation.
procedure TCollection.RetainAll(const Collection: ICollection);
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Enumerator(Enum);
  while Enum.Next(Item) do
    if not Collection.Contains(Item) then
      Enum.Remove;
end;

// Convert the collection into a dynamic array.
function TCollection.ToArray: TCollectibleArray;
var
  Enum: IEnumerator;
begin
  Enumerator(Enum);
  Enum.Next(Count, Result);
  if Length(Result) <> Count then
    raise ECollectionError.Create(sCollectionChanged);
end;

procedure TCollection.DuplicateError(Item: ICollectible);
begin
  raise ECollectionItemError.Create(sDuplicate, [ClassName], Item);
end;

procedure TCollection.ItemNotFound(Item: ICollectible);
begin
  raise ECollectionItemError.Create(sItemNotFound, [ClassName], Item);
end;

procedure TCollection.ValueNotFound(Item: ICollectible);
begin
  raise ECollectionItemError.Create(sValueNotFound, [ClassName], Item);
end;

procedure TCollection.IndexOutOfRange(Index: Integer);
begin
  raise ECollectionError.CreateFmt(sIndexOutOfRange, [ClassName, Index]);
end;

procedure TCollection.NotImplemented(const MethodName: string);
begin
  raise ECollectionError.CreateFmt(sNotImplemented, [ClassName, MethodName]);
end;



{ TEnumerator }

constructor TEnumerator.Create(Collection: TCollection);
begin
  inherited Create;
  fCollection := Collection;
end;

procedure TEnumerator.Clone(out Enum: IEnumerator);
begin
  Enum := TEnumerator(ClassType).Create(CollectionObject);
end;

// Get the next Count items and store them in ResultArray. Set NumFetched
// to the actual number of items fetched. If Count=NumFetched, return S_OK,
// otherwise, return S_False.
//
// This style of Next follows Microsoft's convention for enumerators, but
// it makes sense only with primitive programming languages, such as C.
// Delphi programs will most likely use one of the more useful forms.
procedure TEnumerator.Next(Count: Integer;
  var ResultArray: TStaticCollectibleArray; var NumFetched: Integer);
var
  I: Integer;
  Item: ICollectible;
begin
  for I := 0 to Count-1 do
  begin
    if not Next(Item) then
    begin
      NumFetched := I;
      Exit;
    end;
    ResultArray[I] := Item;
  end;
  NumFetched := Count;
end;

// Get up to Count items and store them in a dynamic array, which is
// resized to accomodate the number of items actually fetched.
type
  PStaticCollectibleArray = ^TStaticCollectibleArray;

procedure TEnumerator.Next(Count: Integer; out ResultArray: TCollectibleArray);
var
  NumFetched: Integer;
begin
  SetLength(ResultArray, Count);
  Next(Count, PStaticCollectibleArray(@ResultArray[1])^, NumFetched);
  if NumFetched <> Count then
    SetLength(ResultArray, NumFetched);
end;

// Skip items without actually fetching them.
function TEnumerator.Skip(Count: Integer): Integer;
var
  I: Integer;
  Item: ICollectible;
begin
  for I := 0 to Count-1 do
    if not Next(Item) then
    begin
      Result := I;
      Exit;
    end;
  Result := Count;
end;

destructor TEnumerator.Destroy;
begin
  fCollection := nil;
  inherited;
end;

procedure TEnumerator.GetCollection(out Collection: ICollection);
begin
  Collection := fCollection;
end;

procedure TEnumerator.CantRemoveError;
begin
  raise ECollectionError.CreateFmt(sCantRemove, [ClassName]);
end;


{ TString }

constructor TString.Create(Value: string);
begin
  inherited Create;
  fValue := Value;
end;

constructor TString.Create(Value: TString);
begin
  Create(Value.Value);
end;

destructor TString.Destroy;
begin
  inherited;
end;

function TString.CompareTo(const Obj: ICollectible): Integer;
begin
  Result := CompareStr(Value, (Obj as IString).Value);
end;

function TString.GetValue: string;
begin
  Result := fValue;
end;

function TString.HashValue: THashValue;
begin
  Result := HashString(Value);
end;

procedure TString.SetValue(const Value: string);
begin
  fValue := Value;
end;

{ TWrapper }

constructor TWrapper.Create(Obj: TObject);
begin
  inherited Create;
  fObject := Obj;
end;

constructor TWrapper.Create(Wrapper: TWrapper);
begin
  Create(Wrapper.Value);
end;

// There is no way, in general, to compare objects, so compare their addresses.
// Return zero only when the two object pointers are the same object.
function TWrapper.CompareTo(const Obj: ICollectible): Integer;
begin
  Result := Ord(Value = (Obj as IObject).Value);
end;

function TWrapper.GetValue: TObject;
begin
  Result := fObject;
end;

// There is no way, in general, to compute a good hash value for an arbitrary
// object, so use the address of the object.
function TWrapper.HashValue: THashValue;
begin
  Result := THashValue(Value);
end;

procedure TWrapper.SetValue(Value: TObject);
begin
  fObject := Value;
end;

{ ECollectionItemError }

constructor ECollectionItemError.Create(const Msg: string; Item: ICollectible);
begin
  inherited Create(Msg);
  fItem := Item;
end;

constructor ECollectionItemError.Create(const Fmt: string; Args: array of const;
  Item: ICollectible);
begin
  inherited CreateFmt(Fmt, Args);
  fItem := Item;
end;

procedure ECollectionItemError.GetItem(out Item: ICollectible);
begin
  Item := fItem;
end;

{ TAssociation }

constructor TAssociation.Create(Key: ICollectible; Value: IUnknown);
begin
  inherited Create;
  fKey := Key;
  fValue := Value;
end;

procedure TAssociation.GetKey(out Key: ICollectible);
begin
  Key := fKey
end;

procedure TAssociation.GetValue(out Value: IUnknown);
begin
  Value := fValue
end;

// Get the hash value of the key, ignoring the value.
function TAssociation.HashValue: THashValue;
begin
  Result := Key.HashValue;
end;

// If the other object is an association, compare the keys;
// otherwise compare this key to the other object itself.
function TAssociation.CompareTo(const Obj: ICollectible): Integer;
var
  Assoc: IAssociation;
  ObjKey: ICollectible;
begin
  if Obj.QueryInterface(IAssociation, Assoc) = S_OK then
  begin
    Assoc.GetKey(ObjKey);
    Result := Key.CompareTo(ObjKey);
  end
  else
    Result := Key.CompareTo(Obj);
end;

procedure TAssociation.SetValue(Value: IUnknown);
begin
  fValue := Value;
end;


{ TReverseWrapper }

constructor TReverseWrapper.Create(Item: ICollectible);
begin
  inherited Create;
  fItem := Item;
end;

// Reverse the direction of the comparison.
function TReverseWrapper.CompareTo(const Obj: ICollectible): Integer;
begin
  Result := -Item.CompareTo(Obj);
end;

// The hash value is unchanged.
function TReverseWrapper.HashValue: THashValue;
begin
  Result := Item.HashValue;
end;

procedure TReverseWrapper.GetItem(out Item: ICollectible);
begin
  Item := fItem;
end;


{ TMap }

procedure TMap.Add(Assoc: IAssociation);
begin
  Collection.Add(Assoc as ICollectible)
end;

procedure TMap.Add(Key: ICollectible; Value: IUnknown);
var
  Assoc: IAssociation;
begin
  Assoc := TAssociation.Create(Key, Value);
  Add(Assoc)
end;

procedure TMap.Add(Obj: ICollectible);
begin
  Add(Obj as IAssociation)
end;

function TMap.ContainsKey(const Key: ICollectible): Boolean;
var
  Assoc: IAssociation;
begin
  Result := FindKey(Key, Assoc);
end;

function TMap.ContainsValue(const Value: ICollectible): Boolean;
var
  Assoc: IAssociation;
begin
  Result := FindValue(Value, Assoc)
end;

constructor TMap.Create;
begin
  inherited;
  CreateCollection(fCollection);
end;

procedure TMap.Enumerator(out Enum: IEnumerator);
begin
  Enum := TEnumMap.Create(Self);
end;

procedure TMap.Enumerator(out Enum: IEnumMap);
begin
  Enum := TEnumMap.Create(Self);
end;

function TMap.FindKey(const Key: ICollectible; out Assoc: IAssociation): Boolean;
var
  Found: ICollectible;
begin
  Result := Find(MakeAssociation(Key) as ICollectible, Found);
  if Result then
    Assoc := Found as IAssociation;
end;

function TMap.FindValue(const Value: ICollectible; out Assoc: IAssociation): Boolean;
var
  Enum: IEnumerator;
  Item: ICollectible;
  AssocValue: IUnknown;
begin
  Enumerator(Enum);
  while Enum.Next(Item) do
  begin
    Assoc := Item as IAssociation;
    Assoc.GetValue(AssocValue);
    if Value.CompareTo(AssocValue as ICollectible) = 0 then
    begin
      Result := True;
      Exit;
    end;
  end;

  Result := False;
end;

procedure TMap.ForEach(Proc: TAssociationProc);
var
  Enum: IEnumerator;
  Item: ICollectible;
  Key: ICollectible;
  Value: IUnknown;
begin
  Enumerator(Enum);
  while Enum.Next(Item) do
    with Item as IAssociation do
    begin
      GetKey(Key);
      GetValue(Value);
      if not Proc(Key, Value) then
        Exit;
    end;
end;

procedure TMap.GetKeys(out Keys: ISet);
var
  Enum: IEnumMap;
  Assoc: IAssociation;
  Key: ICollectible;
begin
  CreateKeysCollection(Keys);
  Enumerator(Enum);
  while Enum.Next(Assoc) do
  begin
    Assoc.GetKey(Key);
    Keys.Add(Key);
  end;
end;

function TMap.GetPropValue(const Key: ICollectible): IUnknown;
begin
  if not GetValue(Key, Result) then
    ItemNotFound(Key)
end;

function TMap.GetValue(const Key: ICollectible; out Value: IUnknown): Boolean;
var
  Assoc: IAssociation;
begin
  Result := FindKey(Key, Assoc);
  if Result then
    Assoc.GetValue(Value);
end;

procedure TMap.GetValues(out Values: ICollection);
var
  Enum: IEnumMap;
  Assoc: IAssociation;
  Value: IUnknown;
begin
  CreateValuesCollection(Values);
  Enumerator(Enum);
  while Enum.Next(Assoc) do
  begin
    Assoc.GetValue(Value);
    Values.Add(Value as ICollectible);
  end;
end;

function TMap.IsMap: Boolean;
begin
  Result := True
end;

procedure TMap.RemoveKey(const Key: ICollectible);
begin
  Remove(MakeAssociation(Key) as ICollectible)
end;

procedure TMap.RemoveValue(const Value: ICollectible);
var
  Enum: IEnumMap;
  Assoc: IAssociation;
  AssocValue: IUnknown;
begin
  Enumerator(Enum);
  while Enum.Next(Assoc) do
  begin
    Assoc.GetValue(AssocValue);
    if Value.CompareTo(AssocValue as ICollectible) = 0 then
    begin
      Enum.Remove;
      Exit;
    end;
  end;
end;

procedure TMap.SetPropValue(const Key: ICollectible; Value: IUnknown);
var
  Assoc: IAssociation;
begin
  if not FindKey(Key, Assoc) then
    ItemNotFound(Key)
  else
    Assoc.SetValue(Value);
end;


{ TEnumMap }

constructor TEnumMap.Create(Collection: TCollection);
begin
  inherited;
  (Collection as TMap).Collection.Enumerator(fEnumerator);
end;

procedure TEnumMap.Clone(out Enum: IEnumerator);
var
  EnumMap: TEnumMap;
begin
  EnumMap := TEnumMap.Create(CollectionObject);
  Enumerator.Clone(EnumMap.fEnumerator);
  Enum := EnumMap;
end;

function TEnumMap.HasNext: Boolean;
begin
  Result := Enumerator.HasNext
end;

function TEnumMap.Next(out Item: ICollectible): Boolean;
begin
  Result := Enumerator.Next(Item);
end;

function TEnumMap.Next(out Obj: IAssociation): Boolean;
var
  Item: ICollectible;
begin
  Result := Next(Item);
  if Result then
    Obj := Item as IAssociation;
end;

// Fetch up to Count items from Enum, storing them in successive
// elements of ResultArray (which must be large enough). Store the number
// of elements actually fetched in NumFetched.
//
// This method follows Microsoft's conventions for COM enumerators, but is
// clumsy to use. Use one of the other methods instead.
procedure TEnumMap.Next(Count: Integer;
  var ResultArray: TStaticAssociationArray;
  var NumFetched: Integer);
var
  I: Integer;
begin
  for I := 0 to Count-1 do
    if not Next(ResultArray[I]) then
    begin
      NumFetched := I;
      Exit;
    end;
  NumFetched := Count;
end;

// Fetch up to Count items from Enum and store them in ResultArray.
// Set the length of ResultArray to the number of items actually fetched.
procedure TEnumMap.Next(Count: Integer; out ResultArray: TAssociationArray);
type
  PStaticAssociationArray = ^TStaticAssociationArray;
var
  NumFetched: Integer;
begin
  SetLength(ResultArray, Count);
  Next(Count, PStaticAssociationArray(@ResultArray[0])^, NumFetched);
  if NumFetched <> Count then
    SetLength(ResultArray, NumFetched);
end;

function TEnumMap.NextKey(out Key: ICollectible): Boolean;
var
  Assoc: IAssociation;
begin
  Result := Next(Assoc);
  if Result then
    Assoc.GetKey(Key)
end;

procedure TEnumMap.Remove;
begin
  Enumerator.Remove;
end;

procedure TEnumMap.Reset;
begin
  Enumerator.Reset;
end;


{ TInteger }

function TInteger.CompareTo(const Obj: ICollectible): Integer;
var
  Int: Integer;
begin
  Int := (Obj as IInteger).Value;
  {$R-}
  Result := Value - Int;
  {$R+}
end;

constructor TInteger.Create(Value: Integer);
begin
  fValue := Value;
end;

function TInteger.GetValue: Integer;
begin
  Result := fValue;
end;

function TInteger.HashValue: THashValue;
begin
  Result := fValue;
end;

procedure TInteger.SetValue(Value: Integer);
begin
  fValue := Value;
end;

{ TDouble }

function TDouble.CompareTo(const Obj: ICollectible): Integer;
var
  Dbl: Double;
begin
  Dbl := (Obj as IDouble).Value;
  if Value < Dbl then
    Result := -1
  else if Value > Dbl then
    Result := 1
  else
    Result := 0;
end;

constructor TDouble.Create(Value: Double);
begin
  fValue := Value;
end;

function TDouble.GetValue: Double;
begin
  Result := fValue;
end;

function TDouble.HashValue: THashValue;
var
  DblAsInt: array[0..1] of Integer;
begin
  Double(DblAsInt) := Value;
  Result := DblAsInt[0] xor DblAsInt[1];
end;

procedure TDouble.SetValue(Value: Double);
begin
  fValue := Value;
end;

{ TSet }

procedure TSet.Add(Obj: ICollectible);
begin
  case Duplicates of
  dupAccept:
    Collection.Add(Obj);
  dupIgnore:
    if not Collection.Contains(Obj) then
      Collection.Add(Obj);
  dupError:
    if Collection.Contains(Obj) then
      DuplicateError(Obj)
    else
      Collection.Add(Obj);
  end;
end;

constructor TSet.Create;
begin
  inherited;
  CreateCollection(fCollection);
end;

procedure TSet.Enumerator(out Enum: IEnumSet);
var
  E: IEnumerator;
begin
  Enumerator(E);
  Enum := E as IEnumSet;
end;

function TSet.GetDuplicates: TDuplicates;
begin
  Result := dupIgnore;
end;

procedure TSet.Intersect(const Collection: ICollection);
begin
  RetainAll(Collection);
end;

procedure TSet.SetDuplicates(Dup: TDuplicates);
begin
  NotImplemented('SetDuplicates');
end;

procedure TSet.Subtract(const Collection: ICollection);
begin
  RemoveAll(Collection);
end;

procedure TSet.Union(const Collection: ICollection);
begin
  AddAll(Collection);
end;

{ TFilterCollection }

procedure TFilterCollection.Add(Obj: ICollectible);
begin
  Collection.Add(Obj);
end;

procedure TFilterCollection.Clear;
begin
  Collection.Clear;
end;

function TFilterCollection.Contains(const Obj: ICollectible): Boolean;
begin
  Result := Collection.Contains(Obj);
end;

procedure TFilterCollection.Enumerator(out Enum: IEnumerator);
begin
  Collection.Enumerator(Enum);
end;

function TFilterCollection.Find(const SearchFor: ICollectible;
  out Found: ICollectible): Boolean;
begin
  Result := Collection.Find(SearchFor, Found);
end;

function TFilterCollection.GetCount: Integer;
begin
  Result := Collection.GetCount;
end;

function TFilterCollection.GetDuplicates: TDuplicates;
begin
  Result := Collection.Duplicates;
end;

function TFilterCollection.IsEmpty: Boolean;
begin
  Result := Collection.IsEmpty;
end;

function TFilterCollection.IsOrdered: Boolean;
begin
  Result := Collection.IsOrdered;
end;

function TFilterCollection.IsSorted: Boolean;
begin
  Result := Collection.IsSorted;
end;

procedure TFilterCollection.Remove(const Obj: ICollectible);
begin
  Collection.Remove(Obj);
end;

procedure TFilterCollection.RemoveAll(const Collection: ICollection);
begin
  Self.Collection.RemoveAll(Collection);
end;

procedure TFilterCollection.RemoveAll(const Obj: ICollectible);
begin
  Collection.RemoveAll(Obj);
end;

procedure TFilterCollection.RetainAll(const Collection: ICollection);
begin
  Self.Collection.RetainAll(Collection);
end;

procedure TFilterCollection.SetDuplicates(Dup: TDuplicates);
begin
  Collection.Duplicates := Dup;
end;

function TFilterCollection.ToArray: TCollectibleArray;
begin
  Result := Collection.ToArray;
end;

{ TInterfaceWrapper }

function TInterfaceWrapper.CompareTo(const Obj: ICollectible): Integer;
begin
  Result := Integer(Intf) - Integer((Obj as IInterface).Intf);
end;

constructor TInterfaceWrapper.Create(Intf: IUnknown);
begin
  inherited Create;
  fInterface := Intf;
end;

function TInterfaceWrapper.GetInterface: IUnknown;
begin
  Result := fInterface;
end;

function TInterfaceWrapper.HashValue: THashValue;
begin
  Result := Integer(Intf);
end;

procedure TInterfaceWrapper.SetInterface(const Value: IUnknown);
begin
  fInterface := Value;
end;

{ TStringItem }

constructor TStringItem.Create(const S: string; Obj: TObject);
begin
  inherited Create(S);
  fObject := Obj;
end;

function TStringItem.GetObject: TObject;
begin
  Result := fObject;
end;

procedure TStringItem.SetObject(Value: TObject);
begin
  fObject := Value;
end;

{ TCltnStrings }

procedure TCltnStrings.Clear;
begin
  Collection.Clear
end;

constructor TCltnStrings.Create;
begin
  inherited;
  CreateCollection(fCollection);
end;

procedure TCltnStrings.CreateCollection(out Collection: IList);
begin
  Collection := TArrayList.Create;
end;

procedure TCltnStrings.CreateItem(out Item: ICollectible; const S: string);
begin
  Item := TStringItem.Create(S);
end;

procedure TCltnStrings.Delete(Index: Integer);
begin
  Collection.Delete(Index);
end;

function TCltnStrings.Get(Index: Integer): string;
begin
  Result := (Collection[Index] as IString).Value;
end;

function TCltnStrings.GetCount: Integer;
begin
  Result := Collection.Count;
end;

function TCltnStrings.GetObject(Index: Integer): TObject;
begin
  Result := (Collection[Index] as IObject).Value;
end;

function TCltnStrings.IndexOf(const S: string): Integer;
begin
  Result := Collection.IndexOf(TString.Create(S));
end;

procedure TCltnStrings.Insert(Index: Integer; const S: string);
var
  Item: ICollectible;
begin
  CreateItem(Item, S);
  Collection.Insert(Index, Item);
end;

procedure TCltnStrings.Put(Index: Integer; const S: string);
begin
  (Collection[Index] as IString).Value := S;
end;

procedure TCltnStrings.PutObject(Index: Integer; AObject: TObject);
begin
  (Collection[Index] as IObject).Value := AObject;
end;

procedure TCltnStrings.Sort;
begin
  Collection.Sort;
end;

procedure TCltnStrings.Sort(Compare: TListCompareFunc);
begin
  Collection.Sort(Compare);
end;

{ TSortedCltnStrings }

procedure TSortedCltnStrings.CreateCollection(out Collection: IList);
begin
  Collection := TSortedList.Create;
end;

// Utilities to help
class function TCltnSort.DefaultCompare(const A, B: ICollectible): Integer;
begin
  Result := A.CompareTo(B);
end;

class procedure TCltnSort.QuickSort(const List: IList; Lo, Hi: Integer;
  Compare: TListCompareFunc);
var
  I, J, Mid: Integer;
begin
  repeat
    I := Lo;
    J := Hi;
    Mid := (Lo + Hi) div 2;
    repeat
      while Compare(List[I], List[Mid]) < 0 do
        Inc(I);
      while Compare(List[J], List[Mid]) > 0 do
        Dec(J);
      if I <= J then
      begin
        List.Exchange(I, J);
        if Mid = I then
          Mid := J
        else if Mid = J then
          Mid := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if Lo < J then
      QuickSort(List, Lo, J, Compare);
    Lo := I;
  until I >= Hi;
end;

class procedure TCltnSort.Sort(const List: IList;
  Compare: TListCompareFunc);
begin
  if List.Count > 0 then
  begin
    if not Assigned(Compare) then
      Compare := DefaultCompare;
    QuickSort(List, 0, List.Count-1, Compare);
  end;
end;

end.
