unit CltnArrays;

(*
  Implement collections using arrays.
  IList and IStack are obvious for array implementations, but ISet and IMap
  can also work with arrays. TSortedList implements IList, but storing all
  items in sorted order. TSortedArraySet and TSortedArrayMap use TSortedList
  to implement ISet and IMap. It turns out a sorted array is only slightly
  slower than TRbTree for many operation, and for collections that don't
  change size much, an array uses less memory than hash tables and trees.
  TArraySet and TArrayMap complete the suite, but are usuable only for
  the tiniest of sets. For sets that contain just a few elements, though,
  the lower overhead may indeed prove beneficial.

  Copyright  1999 Tempest Software, Inc.
*)

interface

uses Windows, SysUtils, Classes, Cltn;

type
  // Simply array collection.
  TArray = class(TCollection)
  private
    fList: TList;
  protected
    property List: TList read fList;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Add(Obj: ICollectible); override;
    procedure Clear; override;
    procedure Delete(Index: Integer);
    procedure Enumerator(out Enum: IEnumerator); override;
    procedure Exchange(I, J: Integer); virtual;
    function Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean; override;
    function GetCount: Integer; override;
    function IndexOf(const Item: ICollectible): Integer; virtual;
    procedure Insert(Index: Integer; Obj: ICollectible); virtual;
    procedure Move(CurIndex, NewIndex: Integer); virtual;
    procedure Remove(const Obj: ICollectible); override;
    procedure RemoveAll(const Obj: ICollectible); override;

    function GetItems(Index: Integer): ICollectible; virtual;
    procedure GetItem(Index: Integer; out Item: ICollectible); virtual;
    procedure SetItems(Index: Integer; Item: ICollectible); virtual;

    property Items[Index: Integer]: ICollectible read GetItems write SetItems; default;
  end;
  TEnumArray = class(TEnumerator)
  private
    fIndex: Integer;
    function GetCollectionList: TArray;
    procedure SetIndex(Index: Integer);
  protected
    property Index: Integer read fIndex write SetIndex;
    property CollectionList: TArray read GetCollectionList;
  public
    constructor Create(Collection: TCollection); override;
    procedure Clone(out Enum: IEnumerator); override;
    function HasNext: Boolean; override;
    function Next(out Item: ICollectible): Boolean; overload; override;
    procedure Remove; override;
    procedure Reset; override;
  end;

  // Implement the list and stack interfaces.
  TArrayList = class(TArray, IList, IStack)
  public
    function IsOrdered: Boolean; override;
    function GetTop(out Top: ICollectible): Boolean;
    procedure Push(Obj: ICollectible);
    function Pop(out Top: ICollectible): Boolean;
    procedure Sort; overload;
    procedure Sort(Compare: TListCompareFunc); overload;
  end;

  // Like an ordered list, but the contents are always in sorted order.
  TSortedList = class(TArray, IList)
  private
    fDuplicates: TDuplicates;
  protected
    function SearchFor(const Item: ICollectible; var Index: Integer): Boolean; virtual;
    function GetDuplicates: TDuplicates; override;
    procedure SetDuplicates(Dup: TDuplicates); override;
  public
    procedure Add(Obj: ICollectible); override;
    function CountMatching(const Obj: ICollectible): Integer; override;
    function IndexOf(const Item: ICollectible): Integer; override;
    procedure Insert(Index: Integer; Obj: ICollectible); override;
    function IsSorted: Boolean; override;
    procedure Sort; overload;
    procedure Sort(Compare: TListCompareFunc); overload;
  end;

  // Array implementation of a set. Use TArraySet only for very small sets.
  TArraySet = class(TSet)
  protected
    procedure CreateCollection(out Collection: ICollection); override;
  end;

  // By keeping the items sorted, add, find, and remove are faster.
  TSortedArraySet = class(TSet)
  protected
    procedure CreateCollection(out Collection: ICollection); override;
  end;
  TSortedArrayMultiSet = class(TSortedArraySet)
  protected
    function GetDuplicates: TDuplicates; override;
    procedure SetDuplicates(Dup: TDuplicates); override;
  end;

  // Array implementation of a map. Use TArrayMap only for very small maps.
  TArrayMap = class(TMap)
  protected
    procedure CreateCollection(out Collection : ICollection); override;
    procedure CreateValuesCollection(out Values: ICollection); override;
    procedure CreateKeysCollection(out Keys: ISet); override;
  public
    function IsOrdered: Boolean; override;
  end;

  TSortedArrayMap = class(TArrayMap)
  protected
    procedure CreateCollection(out Collection : ICollection); override;
    procedure CreateKeysCollection(out Keys: ISet); override;
  public
    function IsSorted: Boolean; override;
  end;

implementation

{ TArray }

constructor TArray.Create;
begin
  inherited;
  fList := TList.Create;
end;

destructor TArray.Destroy;
begin
  Clear;
  List.Free;
  fList := nil;
  inherited;
end;

procedure TArray.Add(Obj: ICollectible);
begin
  Obj._AddRef;
  List.Add(Pointer(Obj));
end;

procedure TArray.Enumerator(out Enum: IEnumerator);
begin
  Enum := TEnumArray.Create(Self);
end;

function TArray.Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean;
var
  Index: Integer;
begin
  Index := IndexOf(SearchFor);
  Result := Index >= 0;
  if Result then
    GetItem(Index, Found);
end;

function TArray.GetCount: Integer;
begin
  Result := List.Count;
end;

procedure TArray.Remove(const Obj: ICollectible);
var
  Index: Integer;
begin
  Index := IndexOf(Obj);
  if Index < 0 then
    ItemNotFound(Obj);
  Delete(Index);
end;

procedure TArray.Delete(Index: Integer);
var
  Item: ICollectible;
begin
  Item := GetItems(Index);
  List.Delete(Index);
  Item._Release;
end;

procedure TArray.Exchange(I, J: Integer);
begin
  List.Exchange(I, J);
end;

procedure TArray.GetItem(Index: Integer; out Item: ICollectible);
begin
  Item := ICollectible(List[Index]);
end;

function TArray.GetItems(Index: Integer): ICollectible;
begin
  Result := ICollectible(List[Index]);
end;

function TArray.IndexOf(const Item: ICollectible): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to List.Count-1 do
    if Item.CompareTo(ICollectible(List[I])) = 0 then
    begin
      Result := I;
      Break;
    end;
end;

procedure TArray.SetItems(Index: Integer; Item: ICollectible);
var
  OldItem: ICollectible;
begin
  OldItem := GetItems(Index);
  Item._AddRef;
  List[Index] := Pointer(Item);
  OldItem._Release;
end;

procedure TArray.Clear;
var
  I: Integer;
begin
  for I := Count-1 downto 0 do
    Delete(I);
end;

procedure TArray.Insert(Index: Integer; Obj: ICollectible);
begin
  Obj._AddRef;
  List.Insert(Index, Pointer(Obj));
end;

procedure TArray.Move(CurIndex, NewIndex: Integer);
begin
  List.Move(CurIndex, NewIndex);
end;

procedure TArray.RemoveAll(const Obj: ICollectible);
var
  I: Integer;
begin
  for I := Count-1 downto 0 do
    if Obj.CompareTo(GetItems(I)) = 0 then
      Delete(I);
end;

{ TArrayList }

function TArrayList.IsOrdered: Boolean;
begin
  Result := True
end;

function TArrayList.GetTop(out Top: ICollectible): Boolean;
begin
  Result := Count > 0;
  if Result then
    GetItem(Count-1, Top);
end;

function TArrayList.Pop(out Top: ICollectible): Boolean;
begin
  Result := GetTop(Top);
  if Result then
    Delete(Count-1);
end;

procedure TArrayList.Push(Obj: ICollectible);
begin
  Add(Obj);
end;

procedure TArrayList.Sort;
begin
  TCltnSort.Sort(Self);
end;

procedure TArrayList.Sort(Compare: TListCompareFunc);
begin
  TCltnSort.Sort(Self, Compare);
end;


{ TEnumArray }

procedure TEnumArray.Clone(out Enum: IEnumerator);
var
  EnumObject: TEnumArray;
begin
  EnumObject := TEnumArray.Create(CollectionObject);
  EnumObject.Index := Index;
  Enum := EnumObject;
end;

// Index points to the next item to fetch, so subtract one
// to get the most recently fetched item.
procedure TEnumArray.Remove;
begin
  if (Index <= 0) or (Index > CollectionList.Count) then
    CantRemoveError;
  Dec(fIndex);
  CollectionList.Delete(Index);
end;

function TEnumArray.Next(out Item: ICollectible): Boolean;
begin
  Result := Index < CollectionList.Count;
  if Result then
  begin
    Item := CollectionList[Index];
    Inc(fIndex);
  end;
end;

// Reset the enumerator to the beginning of the collection.
procedure TEnumArray.Reset;
begin
  Index := 0;
end;

procedure TEnumArray.SetIndex(Index: Integer);
begin
  if (Index < 0) or (Index >= CollectionList.Count) then
    CollectionList.IndexOutOfRange(Index);
  fIndex := Index;
end;

function TEnumArray.GetCollectionList: TArray;
begin
  Result := TArray(CollectionObject);
end;

function TEnumArray.HasNext: Boolean;
begin
  Result := Index < CollectionList.Count;
end;

constructor TEnumArray.Create(Collection: TCollection);
resourcestring
  sInvalidCollection = 'TEnumArray requires TArray collection, not %s';
begin
  if not (Collection is TArray) then
    raise EInvalidCast.CreateFmt(sInvalidCollection, [Collection.ClassName]);
  inherited;
end;


{ TSortedList }

resourcestring
  sNotImplemented = 'Sorted list: Insert method is not implemented';
  
// Add an item at its proper position.
procedure TSortedList.Add(Obj: ICollectible);
var
  Index: Integer;
begin
  if SearchFor(Obj, Index) then
    case Duplicates of
    dupIgnore:
      Exit;
    dupError:
      DuplicateError(Obj);
    else
      {do nothing};
    end;
  Obj._AddRef;
  List.Insert(Index, Pointer(Obj));
end;

// All matching items will be adjacent.
function TSortedList.CountMatching(const Obj: ICollectible): Integer;
var
  Index: Integer;
begin
  Result := 0;
  if SearchFor(Obj, Index) then
    repeat
      Inc(Index);
      Inc(Result);
    until Obj.CompareTo(Items[Index]) <> 0;
end;

// Use a binary search to find an item.
function TSortedList.GetDuplicates: TDuplicates;
begin
  Result := fDuplicates;
end;

function TSortedList.IndexOf(const Item: ICollectible): Integer;
begin
  if not SearchFor(Item, Result) then
    Result := -1;
end;

// Perform a binary search for Item in the list, and set Index to the position
// of the item or where the item should be inserted. Return True if the
// item is found, False if not found.
procedure TSortedList.Insert(Index: Integer; Obj: ICollectible);
begin
  raise ECollectionError.Create(sNotImplemented);
end;

function TSortedList.IsSorted: Boolean;
begin
  Result := True
end;

// Binary search.
function TSortedList.SearchFor(const Item: ICollectible; var Index: Integer): Boolean;
var
  Low, Mid, High: Integer;
  Compare: Integer;
begin
  Low := 0;
  High := List.Count-1;
  Result := False;

  while Low <= High do
  begin
    Mid := (Low + High) div 2;
    Compare := Item.CompareTo(GetItems(Mid));
    if Compare > 0 then
      Low := Mid + 1
    else if Compare < 0 then
      High := Mid - 1
    else
    begin
      Result := True;
      Index := Mid;
      Exit;
    end;
  end;
  Index := Low;
end;

procedure TSortedList.SetDuplicates(Dup: TDuplicates);
begin
  fDuplicates := Dup;
end;


procedure TSortedList.Sort;
begin
  // Already sorted: nothing more to do
end;

procedure TSortedList.Sort(Compare: TListCompareFunc);
begin
  TCltnSort.Sort(Self, Compare);
end;

{ TArraySet }

procedure TArraySet.CreateCollection(out Collection: ICollection);
begin
  Collection := TArray.Create;
end;


{ TArrayMap }

procedure TArrayMap.CreateCollection(out Collection: ICollection);
begin
  Collection := TArraySet.Create
end;

procedure TArrayMap.CreateKeysCollection(out Keys: ISet);
begin
  Keys := TArraySet.Create
end;

procedure TArrayMap.CreateValuesCollection(out Values: ICollection);
begin
  Values := TArrayList.Create
end;

function TArrayMap.IsOrdered: Boolean;
begin
  Result := True
end;

{ TSortedArrayMap }

procedure TSortedArrayMap.CreateCollection(out Collection: ICollection);
begin
  Collection := TSortedArraySet.Create;
end;

procedure TSortedArrayMap.CreateKeysCollection(out Keys: ISet);
begin
  Keys := TSortedArraySet.Create;
end;

function TSortedArrayMap.IsSorted: Boolean;
begin
  Result := True
end;

{ TSortedArraySet }


{ TSortedArraySet }

procedure TSortedArraySet.CreateCollection(out Collection: ICollection);
begin
  Collection := TSortedList.Create;
end;

{ TSortedArrayMultiSet }

function TSortedArrayMultiSet.GetDuplicates: TDuplicates;
begin
  Result := Collection.Duplicates;
end;

procedure TSortedArrayMultiSet.SetDuplicates(Dup: TDuplicates);
begin
  Collection.Duplicates := Dup
end;

end.
