unit CltnHash;

(*
  Hash tables.
  The hash tables implement the ISet and IMap interfaces, using a closed
  hash table as the implementation. Items with the same hash value are
  stored in a dynamic, unordered array.

  A possible performance improvement if you know that the trees
  are truly ordered is to use a TSortedList instead of TArrayList.
  Changing a linear search to a binary search can help speed up
  access if the hash table faces many collisions and is densely packed.
  I haven't done this because a hash table that does not face
  heavy collisions doesn't need the additional overhead. More important
  is that as currently implemented, one can use a hash table to
  store objects that don't really have a defined order. Yes, they
  must implement ICollectible, but sometimes, only a comparison for
  equality is meaningful. In that case, a hash table is more suitable
  than, say, a tree.

  The programmer can set the treshold percentage. When the count/capacity
  ratio reaches the threshold, the hash table grows.

  The hash table implementation uses a TList for the main hash table.
  Each element of the list is either nil or another TList. The sublist
  contains the actual ICollectible items.
  
  Copyright  1999 Tempest Software, Inc.
*)

interface

uses Windows, SysUtils, Classes, Cltn;

type
  TPercent = 0..100;
  THashTable = class(TCollection)
  private
    fCount: Integer;
    fCapacity: Integer;
    fTable: TList;
    fThreshold: TPercent;
  protected
    function Lookup(const Key: ICollectible; var SubList: TList; var Index: Integer): Boolean; virtual;
    function GetCapacity: Integer; virtual;
    procedure SetCapacity(Capacity: Integer); virtual;
    procedure SetThreshold(Threshold: TPercent); virtual;
    function GetCount: Integer; override;
    procedure GrowTable; virtual;
    procedure Rehash; virtual;
    property Table: TList read fTable;
  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Add(Key: ICollectible); overload; override;
    procedure Clear; override;
    procedure Enumerator(out Enum: IEnumerator); override;
    function Find(const Key: ICollectible; out Found: ICollectible): Boolean; override;
    procedure Remove(const Key: ICollectible); override;

    property Capacity: Integer read fCapacity write SetCapacity;
    property Threshold: TPercent read fThreshold write SetThreshold;
  end;

  THashSet = class(TSet)
  private
    fHashTable: THashTable;
  protected
    procedure CreateCollection(out Collection: ICollection); override;
    property HashTable: THashTable read fHashTable;
  public
    procedure Enumerator(out Enum: IEnumerator); override;
  end;

  THashMap = class(TMap)
  private
    fHashTable: THashTable;
  protected
    procedure CreateCollection(out Collection: ICollection); override;
    procedure CreateValuesCollection(out Values: ICollection); override;
    procedure CreateKeysCollection(out Keys: ISet); override;
    property HashTable: THashTable read fHashTable;
  end;

  EHashError = class(Exception)
  end;

  TEnumHashTable = class(TEnumerator)
  private
    fListIndex: Integer;
    fSubIndex: Integer;
    fSubList: IList;
    function GetHashTable: THashTable;
  protected
    constructor Create(Collection: TCollection); override;
    procedure CloneBase(Enum: TEnumHashTable);
    property HashTable: THashTable read GetHashTable;
  public
    function HasNext: Boolean; override;
    function Next(out Item: ICollectible): Boolean; override;
    procedure Remove; override;
    procedure Reset; override;

    property ListIndex: Integer read fListIndex write fListIndex;
    property SubIndex: Integer read fSubIndex write fSubIndex;
    property SubList: IList read fSubList write fSubList;
  end;
  TEnumHashSet = class(TEnumHashTable, IEnumSet)
  public
    procedure Clone(out Enum: IEnumerator); override;
  end;
  TEnumHashMap = class(TEnumMap)
  end;

const
  InitialCapacity = 37;  // A small prime number is a good starting size.
  InitialThreshold = 67; // Expand the hash table when it reaches 67% full.

implementation

uses CltnArrays;

resourcestring
  sInvalidCollection = 'TEnumHashTable requires THashTable collection, not %s';

{ THashTable }

constructor THashTable.Create;
begin
  inherited;
  fThreshold := InitialThreshold;
  fTable := TList.Create;
  Capacity := InitialCapacity;
end;

destructor THashTable.Destroy;
begin
  if Table <> nil then
    Clear;
  Table.Free;
  fTable := nil;
  inherited;
end;

// Add an item to the table. If the main table's entry is nil,
// create a sub list; otherwise get the sublis that already exists.
// If the sublist exists, check to see whether the item is already
// in the sublist.
// Add the item to the sublist.
procedure THashTable.Add(Key: ICollectible);
var
  SubList: TList;
  I: Integer;
  Hash: Integer;
begin
  Hash := Key.HashValue mod Capacity;
  if Table[Hash] = nil then
    Table[Hash] := TList.Create;
  SubList := Table[Hash];

  for I := 0 to SubList.Count-1 do
    if Key.CompareTo(ICollectible(SubList[I])) = 0 then
      case Duplicates of
      dupIgnore:
        Exit;
      dupError:
        DuplicateError(Key);
      else
        {do nothing};
      end;

  // Add the new key to the sublist.
  Key._AddRef;
  SubList.Add(Pointer(Key));
  fCount := Count + 1;

  // If the table has too many items, expand the table to avoid hash collisions.
  if Count > MulDiv(Capacity, Threshold, 100) then
    GrowTable;
end;

// Clear the contents of the hash table and free all the sublists.
// Reset the table back to its default capacity so it is ready to
// accept more items.
procedure THashTable.Clear;
var
  I, J: Integer;
  SubList: TList;
  Key: ICollectible;
begin
  for I := 0 to Table.Count-1 do
    if Table[I] <> nil then
    begin
      SubList := Table[I];
      Table[I] := nil;
      for J := 0 to SubList.Count-1 do
      begin
        Key := ICollectible(SubList[J]);
        SubList[J] := nil;
        Key._Release;
        Dec(fCount);
      end;
      SubList.Free;
    end;
  Capacity := InitialCapacity;
  Assert(Count = 0);
end;

// Remove an item from the table by removing it from its sublist.
// If the sublist becomes empty, leave it in the main table
// so a subsequent Add can add it to the sublist without
// creating a new sublist.
procedure THashTable.Remove(const Key: ICollectible);
var
  SubList: TList;
  Index: Integer;
  Item: ICollectible;
begin
  if not Lookup(Key, SubList, Index) then
    ItemNotFound(Key);
  Item := ICollectible(SubList[Index]);
  Sublist.Delete(Index);
  Item._Release;
  Dec(fCount);
end;

function THashTable.Find(const Key: ICollectible; out Found: ICollectible): Boolean;
var
  SubList: TList;
  Index: Integer;
begin
  Result := Lookup(Key, SubList, Index);
  if Result then
    Found := ICollectible(SubList[Index]);
end;

function THashTable.GetCapacity: Integer;
begin
  Result := fCapacity;
end;

function THashTable.GetCount: Integer;
begin
  Result := fCount;
end;

// Lookup Key in the hash table. Find the sublist where the key belongs,
// and its index in the sublist. Return True if the key is at that position,
// False if the key is not in the sublist.
function THashTable.Lookup(const Key: ICollectible; var SubList: TList;
  var Index: Integer): Boolean;
var
  I: Integer;
  Hash: Integer;
begin
  Hash := Key.HashValue mod Capacity;
  SubList := Table[Hash];
  if SubList <> nil then
    for I := 0 to SubList.Count-1 do
      if Key.CompareTo(ICollectible(SubList[I])) = 0 then
      begin
        Index := I;
        Result := True;
        Exit;
      end;
  Result := False;
end;

// Rehash the table after changing its capacity. Save a reference to the old
// table and create a new one with the new capacity. Move each item from the
// old table to the new one. Free the old table and all of its sublists.
procedure THashTable.Rehash;
var
  OldTable: TList;
  I, J: Integer;
  SubList: TList;
  Item: ICollectible;
begin
  OldTable := Table;
  try
    fTable := TList.Create;
    fCount := 0;
    Table.Count := Capacity;
    for I := 0 to OldTable.Count-1 do
    begin
      SubList := OldTable[I];
      OldTable[I] := nil;
      if SubList <> nil then
        for J := 0 to SubList.Count-1 do
        begin
          Item := ICollectible(SubList[J]);
          Add(Item);
          SubList[J] := nil;
          Item._Release;
        end;
      SubList.Free;
    end;
  finally
    OldTable.Free;
  end;
end;

// Set the table capacity, but only if the new capacity is large
// enough to hold the items currently stored in the Table.
// This might be an unwarranted intrusion, but I fail to see
// why one should be allowed to create an overly crowded hash table.
procedure THashTable.SetCapacity(Capacity: Integer);
begin
  if Capacity >= Count then
  begin
    fCapacity := Capacity;
    Rehash;
  end;
end;

// Set the threshold. If the new threshold is too small, immediately
// rehash the table.
procedure THashTable.SetThreshold(Threshold: TPercent);
begin
  fThreshold := Threshold;
  if (Count * Threshold) div 100 > Capacity then
    Rehash;
end;

// Make the table bigger. Setting the capacity automatically
// allocates a new Table and copies the items from the old table
// to the new one.
procedure THashTable.GrowTable;
begin
  Capacity := Capacity * 2 + 5
end;



{ TEnumHashTable }

// Base class method to finish a Clone operation: copy the fields
// from Self to Enum (the clone).
procedure TEnumHashTable.CloneBase(Enum: TEnumHashTable);
begin
  Enum.ListIndex := ListIndex;
  Enum.SubIndex := SubIndex;
  Enum.SubList := SubList;
end;

constructor TEnumHashTable.Create(Collection: TCollection);
begin
  if not (Collection is THashTable) then
    raise EInvalidCast.CreateFmt(sInvalidCollection, [Collection.ClassName]);
  inherited;
  ListIndex := 0;
  SubIndex := 0;
  SubList := nil;
end;

function TEnumHashTable.GetHashTable: THashTable;
begin
  Result := THashTable(CollectionObject);
end;

function TEnumHashTable.HasNext: Boolean;
var
  SubList: TList;
  I: Integer;
  SubIdx: Integer;
begin
  Result := False;
  SubIdx := SubIndex;
  for I := ListIndex to HashTable.Table.Count-1 do
  begin
    SubList := HashTable.Table[I];
    if (SubList <> nil) and (SubIdx < Sublist.Count) then
    begin
      Result := True;
      Exit;
    end;
    SubIdx := 0;
  end;
end;

// Find the next item in the hash table. Get the next item in the current
// sublist, and when the sublist is exhausted, find the next sublist.
// Return False when reaching the end of the table.
function TEnumHashTable.Next(out Item: ICollectible): Boolean;
var
  SubList: TList;
  I: Integer;
begin
  for I := ListIndex to HashTable.Table.Count-1 do
  begin
    SubList := HashTable.Table[I];
    if (SubList <> nil) and (SubIndex < SubList.Count) then
    begin
      Result := True;
      Item := ICollectible(SubList[SubIndex]);
      ListIndex := I;
      Inc(fSubIndex);
      Exit;
    end;
    SubIndex := 0;
  end;
  ListIndex := HashTable.Count;
  Result := False;
end;

procedure TEnumHashTable.Remove;
var
  SubList: TList;
  Item: ICollectible;
begin
  if ListIndex >= HashTable.Table.Count then
    CantRemoveError;
  SubList := HashTable.Table[ListIndex];
  if SubList = nil then
    CantRemoveError;

  // The SubIndex is 1 past the item that was last fetched by Next.
  if (SubIndex <= 0) or (SubIndex > SubList.Count) then
    CantRemoveError;

  Dec(fSubIndex);
  Item := ICollectible(SubList[SubIndex]);
  SubList.Delete(SubIndex);
  Item._Release;
  Dec(HashTable.fCount);
end;

// Reset the enumerator to the beginning of the collection.
procedure TEnumHashTable.Reset;
begin
  fListIndex := 0;
  fSubIndex := 0;
end;

procedure THashTable.Enumerator(out Enum: IEnumerator);
begin
  Enum := TEnumHashTable.Create(Self);
end;


{ THashSet }

procedure THashSet.CreateCollection(out Collection: ICollection);
begin
  fHashTable := THashTable.Create;
  Collection := fHashTable;
end;

procedure THashSet.Enumerator(out Enum: IEnumerator);
begin
  Enum := TEnumHashSet.Create(HashTable);
end;


{ TEnumHashSet }

procedure TEnumHashSet.Clone(out Enum: IEnumerator);
var
  EnumObject: TEnumHashSet;
begin
  EnumObject := TEnumHashSet.Create(HashTable);
  CloneBase(EnumObject);
  Enum := EnumObject;
end;


{ THashMap }

procedure THashMap.CreateCollection(out Collection: ICollection);
begin
  fHashTable := THashTable.Create;
  Collection := fHashTable;
end;

procedure THashMap.CreateKeysCollection(out Keys: ISet);
begin
  Keys := THashSet.Create;
end;

procedure THashMap.CreateValuesCollection(out Values: ICollection);
begin
  Values := TArrayList.Create;
end;

end.
