unit CltnList;

(*
  Implement the IList and IStack collections using doubly-linked lists.

  The TLinkedList class keeps track of the head and tail nodes and a
  count of the number of nodes in the list. It's all straightforward stuff.

  There's no reason why one could not implement ISet and IMap or
  create a sorted linked list, but these classes are left as
  an exercise for the reader.

  Copyright  1999 Tempest Software, Inc.
*)

interface

uses Windows, SysUtils, Classes, Cltn;

type
  // Node in a linked list.
  TListNode = class
  private
    fNext, fPrevious: TListNode;
    fItem: ICollectible;
    property Item: ICollectible read fItem write fItem;
  protected
    constructor Create(Item: ICollectible; Previous: TListNode = nil; Next: TListNode = nil);
    property Next: TListNode read fNext write fNext;
    property Previous: TListNode read fPrevious write fPrevious;
  end;

  TLinkedList = class(TCollection, IList, IStack)
  private
    fHead: TListNode;
    fTail: TListNode;
    fCount: Integer;
  protected
    property Head: TListNode read fHead write fHead;
    property Tail: TListNode read fTail write fTail;
  public
    destructor Destroy; override;
    procedure Add(Obj: ICollectible); override;
    procedure Delete(Index: Integer);
    procedure Clear; override;
    procedure Enumerator(out Enum: IEnumerator); override;
    procedure Exchange(I, J: Integer); virtual;
    function Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean; override;
    procedure GetItem(Index: Integer; out Item: ICollectible); virtual;
    function IndexOf(const Obj: ICollectible): Integer;
    procedure Insert(Index: Integer; Obj: ICollectible); virtual;
    function IsOrdered: Boolean; override;
    function GetCount: Integer; override;
    procedure Move(CurIndex, NewIndex: Integer);
    procedure Remove(const Obj: ICollectible); override;
    procedure RemoveAll(const Obj: ICollectible); override;

    procedure Sort; overload;
    procedure Sort(Compare: TListCompareFunc); overload;

    // IStack
    function GetTop(out Top: ICollectible): Boolean;
    procedure Push(Obj: ICollectible);
    function Pop(out Top: ICollectible): Boolean;

    function GetItems(Index: Integer): ICollectible;
    procedure SetItems(Index: Integer; Item: ICollectible);
    property Items[Index: Integer]: ICollectible read GetItems write SetItems; default;
  end;

  TEnumLinkedList = class(TEnumerator, IEnumList, IEnumerator)
  private
    fNode: TListNode;
    function GetLinkedList: TLinkedList;
  protected
    property Node: TListNode read fNode write fNode;
    property LinkedList: TLinkedList read GetLinkedList;
  public
    constructor Create(Collection: TCollection); override;
    procedure Clone(out Enum: IEnumerator); override;
    procedure Remove; override;
    function HasNext: Boolean; override;
    procedure Insert(Obj: ICollectible);
    function Next(out Item: ICollectible): Boolean; overload; override;
    procedure Reset; override;
  end;

implementation

{ TLinkedList }

procedure TLinkedList.Add(Obj: ICollectible);
var
  Node: TListNode;
begin
  if Head = nil then
  begin
    // The new node is the sole item in the list.
    Head := TListNode.Create(Obj);
    Tail := Head;
  end
  else
  begin
    // Add the new Node to the end of the list.
    Node := TListNode.Create(Obj, Tail);
    Tail.Next := Node;
    Tail := Node;
  end;
  Inc(fCount);
end;

procedure TLinkedList.Clear;
var
  Node: TListNode;
  Next: TListNode;
begin
  Node := Head;
  while Node <> nil do
  begin
    Next := Node.Next;
    Node.Free;
    Node := Next;
    Dec(fCount);
  end;
  Head := nil;
  Tail := nil;
  Assert(Count = 0);
end;

procedure TLinkedList.Delete(Index: Integer);
var
  Enum: IEnumerator;
begin
  Enumerator(Enum);
  Enum.Skip(Index);
  Enum.Remove;
end;

destructor TLinkedList.Destroy;
begin
  Clear;
  inherited;
end;

procedure TLinkedList.Enumerator(out Enum: IEnumerator);
begin
  Enum := TEnumLinkedList.Create(Self);
end;

procedure TLinkedList.Exchange(I, J: Integer);
begin
  if I < J then
  begin
    Move(I, J);
    Move(J, I);
  end
  else
  begin
    Move(J, I);
    Move(I, J);
  end;
end;

function TLinkedList.Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean;
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Enumerator(Enum);
  while Enum.Next(Item) do
    if SearchFor.CompareTo(Item) = 0 then
    begin
      Result := True;
      Found := Item;
      Exit;
    end;

  Result := False;
end;

function TLinkedList.GetCount: Integer;
begin
  Result := fCount;
end;

procedure TLinkedList.GetItem(Index: Integer; out Item: ICollectible);
var
  Enum: IEnumerator;
begin
  Enumerator(Enum);
  if (Enum.Skip(Index) < Index) or not Enum.Next(Item) then
    IndexOutOfRange(Index);
end;

function TLinkedList.GetItems(Index: Integer): ICollectible;
begin
  GetItem(Index, Result);
end;

function TLinkedList.GetTop(out Top: ICollectible): Boolean;
begin
  Result := Head <> nil;
  if Result then
    Top := Head.Item;
end;

function TLinkedList.IndexOf(const Obj: ICollectible): Integer;
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Enumerator(Enum);
  Result := 0;
  while Enum.Next(Item) do
  begin
    if Obj.CompareTo(Item) = 0 then
      Exit;
    Inc(Result);
  end;
  Result := -1;
end;

procedure TLinkedList.Insert(Index: Integer; Obj: ICollectible);
var
  Enum: TEnumLinkedList;
  EnumItem: ICollectible;
  I: Integer;
begin
  Enum := TEnumLinkedList.Create(Self);
  for I := 0 to Index-1 do
    if not Enum.Next(EnumItem) then
      IndexOutOfRange(Index);
  Enum.Insert(Obj);
end;

function TLinkedList.IsOrdered: Boolean;
begin
  Result := True;
end;

procedure TLinkedList.Move(CurIndex, NewIndex: Integer);
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  if CurIndex <> NewIndex then
  begin
    Enumerator(Enum);
    if (Enum.Skip(CurIndex) <> CurIndex) or not Enum.Next(Item) then
      IndexOutOfRange(CurIndex);
    Enum.Remove;
    Insert(NewIndex, Item);
  end;
end;

function TLinkedList.Pop(out Top: ICollectible): Boolean;
begin
  Result := GetTop(Top);
  if Result then
    Delete(0);
end;

procedure TLinkedList.Push(Obj: ICollectible);
begin
  Insert(0, Obj);
end;

procedure TLinkedList.Remove(const Obj: ICollectible);
var
  Enum: IEnumerator;
  Item: ICollectible;
begin
  Enumerator(Enum);
  while Enum.Next(Item) do
    if Obj.CompareTo(Item) = 0 then
    begin
      Enum.Remove;
      Exit;
    end;

  ItemNotFound(Obj);
end;

procedure TLinkedList.RemoveAll(const Obj: ICollectible);
var
  Enum: TEnumLinkedList;
  Item: ICollectible;
begin
  Enum := TEnumLinkedList.Create(Self);
  while Enum.Next(Item) do
    if Obj.CompareTo(Item) = 0 then
      Enum.Remove;
end;

procedure TLinkedList.SetItems(Index: Integer; Item: ICollectible);
var
  Enum: TEnumLinkedList;
  EnumItem: ICollectible;
  I: Integer;
begin
  Enum := TEnumLinkedList.Create(Self);
  for I := 0 to Index-1 do
    if not Enum.Next(EnumItem) then
      IndexOutOfRange(Index);
  Enum.Node.Item := Item;
end;

{ TEnumLinkedList }

// The enumerator's Node field points to the current node. The special
// value, Finished, means the enumerator has run to the end of the list.
// Don't use nil because that means the enumerator has not yet started
// to scan the list.
const
  Finished = TListNode($FFFFFFFF);

procedure TEnumLinkedList.Clone(out Enum: IEnumerator);
var
  EnumObject: TEnumLinkedList;
begin
  EnumObject := TEnumLinkedList.Create(CollectionObject);
  EnumObject.Node := Node;
  Enum := EnumObject;
end;

constructor TEnumLinkedList.Create(Collection: TCollection);
resourcestring
  sInvalidCollection = 'TEnumLinkedList requires TLinkedList collection, not %s';
begin
  if not (Collection is TLinkedList) then
    raise EInvalidCast.CreateFmt(sInvalidCollection, [Collection.ClassName]);
  inherited;
end;

procedure TEnumLinkedList.Remove;
var
  NodeToDelete: TListNode;
begin
  // If the enumerator has not started yet, position it on the first node.
  if Node = nil then
    CantRemoveError;

  if Node = Finished then
    CantRemoveError;
  NodeToDelete := Node;
  Node := Node.Previous; // so Next fetches the next item

  if NodeToDelete.Previous = nil then
    LinkedList.Head := NodeToDelete.Next
  else
    NodeToDelete.Previous.Next := NodeToDelete.Next;
  if NodeToDelete.Next = nil then
    LinkedList.Tail := NodeToDelete.Previous
  else
    NodeToDelete.Next.Previous := NodeToDelete.Previous;
  NodeToDelete.Free;
  Dec(LinkedList.fCount);
end;

function TEnumLinkedList.GetLinkedList: TLinkedList;
begin
  Result := TLinkedList(CollectionObject);
end;

function TEnumLinkedList.HasNext: Boolean;
begin
  if Node = Finished then
    Result := False
  else if Node <> nil then
    Result := Node.Next <> nil
  else
    Result := LinkedList.Head <> nil;
end;

// Insert Obj after the current node.
procedure TEnumLinkedList.Insert(Obj: ICollectible);
var
  NewNode: TListNode;
begin
  NewNode := TListNode.Create(Obj);
  if Node = nil then
  begin
    NewNode.Next := LinkedList.Head;
    LinkedList.Head := NewNode;
    if LinkedList.Tail = nil then
      LinkedList.Tail := NewNode;
    Node := NewNode;
  end
  else if Node = Finished then
  begin
    NewNode.Previous := LinkedList.Tail;
    LinkedList.Tail := NewNode;
    if LinkedList.Head = nil then
      LinkedList.Head := NewNode;
    Node := NewNode;
  end
  else
  begin
    NewNode.Next := Node.Next;
    NewNode.Previous := Node;
    if Node.Next = nil then
      LinkedList.Tail := NewNode
    else
      Node.Next.Previous := NewNode;
    Node.Next := NewNode;
    Node := NewNode;
  end;
end;

function TEnumLinkedList.Next(out Item: ICollectible): Boolean;
begin
  // Advance to the next node in the list.
  if Node = nil then
    Node := LinkedList.Head
  else
    Node := Node.Next;

  // If we reached the end of the list, remember that fact, and return nil.
  Result := Node <> nil;
  if Result then
    Item := Node.Item
  else
    Node := Finished;
end;

// Reset the enumerator to the beginning of the collection.
procedure TEnumLinkedList.Reset;
begin
  Node := nil;
end;


procedure TLinkedList.Sort;
begin
  NotImplemented('Sort');
end;

procedure TLinkedList.Sort(Compare: TListCompareFunc);
begin
  NotImplemented('Sort');
end;

{ TListNode }

constructor TListNode.Create(Item: ICollectible; Previous, Next: TListNode);
begin
  inherited Create;
  fNext := Next;
  fPrevious := Previous;
  fItem := Item;
end;


end.
