unit CltnRbTree;

(*
  Red-black trees.
  The trees implement the ISet and IMap interfaces, using red-black trees
  as the implementation. See Introduction to Algorithms, by Cormen, Leiserson,
  and Rivest for details.

  The "color" of each node is stored by setting or clearing the
  least-significant bit in the node's parent pointer. This trick works
  because the LSBit is clear for word-aligned pointers. Delphi always allocates
  memory on aligned boundaries, so the least significant bit is always zero.

  Note that saving 4 bytes of memory improves performance and more than
  compensates for the additional time needed to access the parent
  pointer and color. Performance tests on an AMD K6-2/266 show that
  using this trick is about 10% faster than using simple, direct
  properties for parent and color. If you don't believe me, find out
  for yourself by defining the "DIRECT" compiler macro.

  Copyright  1999 Tempest Software, Inc.
*)

(* Use the following directive to use the simple, direct
   approach to implementing Red-Black trees. In this particular
   case, simplicity is slower than being tricky.
{ $define DIRECT}
*)

interface

uses Windows, SysUtils, Classes, Cltn;

type
  TRbColor = (rbBlack, rbRed);
  TRbNode = class
  private
    fLeft, fRight: TRbNode;
    fItem: ICollectible;
{$ifdef DIRECT}
    fColor: TRbColor;
    fParent: TRbNode;
{$else}
    fParent: Integer;
{$endif}
{$ifndef DIRECT}
    function GetParent: TRbNode;
    procedure SetParent(Parent: TRbNode);
{$endif}
  protected
    constructor Create(Item: ICollectible; Leaf, Parent: TRbNode; Color: TRbColor);
    function IsRed: Boolean;
    function IsBlack: Boolean;
{$ifndef DIRECT}
    function GetColor: TRbColor; virtual;
    procedure SetColor(Color: TRbColor); virtual;
{$endif}

    property Left: TRbNode read fLeft write fLeft;
    property Right: TRbNode read fRight write fRight;
    property Item: ICollectible read fItem write fItem;
{$ifdef DIRECT}
    property Color: TRbColor read fColor write fColor;
    property Parent: TRbNode read fParent write fParent;
{$else}
    property Color: TRbColor read GetColor write SetColor;
    property Parent: TRbNode read GetParent write SetParent;
{$endif}
  end;

  TRbTree = class(TCollection)
  private
    fRoot: TRbNode;
    fLeaf: TRbNode;
    fCount: Integer;
  protected
    procedure LeftRotate(Node: TRbNode);
    procedure RightRotate(Node: TRbNode);
    function FindNode(const Obj: ICollectible): TRbNode;
    procedure RemoveNode(Node: TRbNode);
    procedure DeleteFixup(Node: TRbNode);
    function Succ(Node: TRbNode): TRbNode;
    function Pred(Node: TRbNode): TRbNode;
    function GetDuplicates: TDuplicates; override;
    property Leaf: TRbNode read fLeaf;
    property Root: TRbNode read fRoot write fRoot;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Add(Obj: ICollectible); overload; override;
    procedure Clear; override;
    procedure Enumerator(out Enum: IEnumerator); override;
    function Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean; override;
    function GetCount: Integer; override;
    function IsSorted: Boolean; override;
    procedure Remove(const Obj: ICollectible); override;

    property Count: Integer read fCount;
  end;

  TEnumRbTree = class(TEnumerator)
  private
    fNode: TRbNode;
    function GetRbTree: TRbTree;
  protected
    property Node: TRbNode read fNode write fNode;
    property RbTree: TRbTree read GetRbTree;
  public
    constructor Create(Collection: TCollection); override;
    procedure Clone(out Enum: IEnumerator); overload; override;
    function HasNext: Boolean; override;
    procedure Remove; override;
    function Next(out Item: ICollectible): Boolean; overload; override;
    procedure Reset; override;
  end;

  TRbTreeSet = class(TSet)
  private
    fTree: TRbTree;
  protected
    procedure CreateCollection(out Collection: ICollection); override;
    property Tree: TRbTree read fTree;
  end;

  TRbTreeMap = class(TMap)
  private
    fTree: TRbTree;
  protected
    procedure CreateCollection(out Collection: ICollection); override;
    procedure CreateValuesCollection(out Values: ICollection); override;
    procedure CreateKeysCollection(out Keys: ISet); override;
    property Tree: TRbTree read fTree;
  end;


  TEnumRbTreeSet = class(TEnumRbTree, IEnumSet)
  end;

  TEnumRbTreeMap = class(TEnumMap)
  end;

implementation

uses CltnArrays;

resourcestring
  sInvalidCollection = 'TEnumRbTree requires TRbTree collection, not %s';

type
  // Leaf nodes in the tree are represented by Leaf nodes, not true nil
  // pointers. The Leaf nodes are always black, and the tree can store
  // arbitrary other values in the other fields. To present re-entrancy,
  // each tree allocates its own Leaf object.
  TRbLeaf = class(TRbNode)
  protected
    constructor Create;
{$ifndef DIRECT}
    function GetColor: TRbColor; override;
{$endif}
    procedure SetColor(Color: TRbColor); override;
  end;


{ TRbNode }

{$ifndef DIRECT}
// Store the node's color as the least-significant bit in the Parent pointer.
// A 1-bit is red and zero is black. Test the color by testing whether the
// LSBit of the fParent field.

const
  ColorMask = $00000001;
  PointerMask = $FFFFFFFE;
{$endif}

constructor TRbNode.Create(Item: ICollectible; Leaf, Parent: TRbNode; Color: TRbColor);
begin
  inherited Create;
  fItem := Item;
{$ifdef DIRECT}
  fParent := Parent;
{$else}
  fParent := Integer(Parent);
  // Start out with color = rbBlack
{$endif}
  fLeft := Leaf;
  fRight := Leaf;
  Self.Color := Color;
end;

{$ifndef DIRECT}
function TRbNode.GetColor: TRbColor;
begin
  if (fParent and ColorMask) <> 0 then
    Result := rbRed
  else
    Result := rbBlack;
end;

function TRbNode.GetParent: TRbNode;
begin
  Result := TRbNode(fParent and PointerMask)
end;
{$endif}

function TRbNode.IsBlack: Boolean;
begin
  Result := Color = rbBlack;
end;

function TRbNode.IsRed: Boolean;
begin
  Result := Color = rbRed;
end;

{$ifndef DIRECT}
procedure TRbNode.SetColor(Color: TRbColor);
begin
  if Color = rbRed then
    fParent := fParent or ColorMask
  else
    fParent := fParent and PointerMask;
end;

procedure TRbNode.SetParent(Parent: TRbNode);
begin
  fParent := (fParent and ColorMask) or Integer(Parent);
end;
{$endif}


{ TRbTree }

constructor TRbTree.Create;
begin
  inherited;
  fLeaf := TRbLeaf.Create;
  Root := Leaf;
end;

destructor TRbTree.Destroy;
begin
  if Root <> nil then
    Clear;
  Leaf.Free;
  fLeaf := nil;
  inherited;
end;

procedure TRbTree.Add(Obj: ICollectible);
var
  Node: TRbNode;
  ParentNode: TRbNode;
  Temp: TRbNode;
  Compare: Integer;
begin
  // Start by finding the right place to insert the new node.
  ParentNode := Leaf;
  Node := Root;
  while Node <> Leaf do
  begin
    ParentNode := Node;
    Compare := Obj.CompareTo(Node.Item);
    if Compare < 0 then
      Node := Node.Left
    else if (Compare > 0) or (Duplicates = dupAccept) then
      Node := Node.Right
    else if Duplicates = dupError then
      DuplicateError(Obj)
    else // Duplicates = dupIgnore
      Exit;
  end;
  // Create the new node with the proper parent, and update
  // the parent's appropriate child field.
  Node := TRbNode.Create(Obj, Leaf, ParentNode, rbRed);
  if ParentNode = Leaf then
    Root := Node
  else if Obj.CompareTo(ParentNode.Item) < 0 then
    ParentNode.Left := Node
  else
    ParentNode.Right := Node;

  // Now rebalance the tree to preserve the red-black properties.
  while (Node <> Root) and Node.Parent.IsRed do
  begin
    Assert(Node.Parent.Parent <> Leaf);
    Assert(Node.Parent.Parent.IsBlack);
    if Node.Parent = Node.Parent.Parent.Left then
    begin
      // The new node's parent is the left child of its parent.
      Temp := Node.Parent.Parent.Right;
      if Temp.IsRed then
      begin
        Node.Parent.Color := rbBlack;
        Temp.Color := rbBlack;
        Node.Parent.Parent.Color := rbRed;
        Node := Node.Parent.Parent;
      end
      else
      begin
        if Node.Parent.Right = Node then
        begin
          Node := Node.Parent;
          LeftRotate(Node);
        end;
        Node.Parent.Color := rbBlack;
        Node.Parent.Parent.Color := rbRed;
        RightRotate(Node.Parent.Parent);
      end;
    end
    else
    begin
      // The new node's parent is the right child of its parent.
      Temp := Node.Parent.Parent.Left;
      if Temp.IsRed then
      begin
        Node.Parent.Color := rbBlack;
        Temp.Color := rbBlack;
        Node.Parent.Parent.Color := rbRed;
        Node := Node.Parent.Parent;
      end
      else
      begin
        if Node.Parent.Left = Node then
        begin
          Node := Node.Parent;
          RightRotate(Node);
        end;
        Node.Parent.Color := rbBlack;
        Node.Parent.Parent.Color := rbRed;
        LeftRotate(Node.Parent.Parent);
      end;
    end;
  end;
  Root.Color := rbBlack;
  Inc(fCount);
end;

procedure TRbTree.Enumerator(out Enum: IEnumerator);
begin
  Enum := TEnumRbTree.Create(Self);
end;

function TRbTree.FindNode(const Obj: ICollectible): TRbNode;
var
  Compare: Integer;
begin
  Result := Root;
  while Result <> Leaf do
  begin
    Compare := Obj.CompareTo(Result.Item);
    if Compare < 0 then
      Result := Result.Left
    else if Compare > 0 then
      Result := Result.Right
    else
      Exit;
  end;
end;

function TRbTree.Find(const SearchFor: ICollectible; out Found: ICollectible): Boolean;
var
  Node: TRbNode;
begin
  Node := FindNode(SearchFor);
  Result := Node <> Leaf;
  if Result then
    Found := Node.Item;
end;

function TRbTree.GetCount: Integer;
begin
  Result := fCount;
end;

procedure TRbTree.Remove(const Obj: ICollectible);
var
  Node: TRbNode;
begin
  Node := FindNode(Obj);
  if Node = Leaf then
    ItemNotFound(Obj)
  else
    RemoveNode(Node);
end;

procedure TRbTree.RemoveNode(Node: TRbNode);
var
  Delete: TRbNode;   // Node that is actually removed from the tree
  Child: TRbNode;    // Non-nil (non-Leaf) child of PhysicalDelete
begin
  // Remove the Delete node from the tree. If Node does not have
  // two children, Node is the one removed, and its child takes its place.
  // Otherwise, find Node's successor to remove. The successor is
  // guaranteed to have at most most one child.
  if (Node.Left = Leaf) or (Node.Right = Leaf) then
    Delete := Node
  else
    Delete := Succ(Node);

  // Set X to the non-nil (that is, non-Leaf) child of Y, if there is one.
  if Delete.Left <> Leaf then
    Child := Delete.Left
  else
    Child := Delete.Right;

  // Change the pointers to remove node Delete from the tree.
  Child.Parent := Delete.Parent;
  if Delete.Parent = Leaf then
    Root := Child
  else if Delete = Delete.Parent.Left then
    Delete.Parent.Left := Child
  else
    Delete.Parent.Right := Child;

  // If Delete is not the same node as Node, copy the Item
  // reference from Delete to Node to save the Item before freeing
  // the Delete node.
  if Delete <> Node then
    Node.Item := Delete.Item;

  // If deleting a black node, the red-black properties are no longer
  // true, so fixup the colors and rebalance the tree.
  if Delete.IsBlack then
    DeleteFixup(Child);

  Delete.Free;
  Dec(fCount);
end;

// Fixup the colors to preserve the red-black properties.
procedure TRbTree.DeleteFixup(Node: TRbNode);
var
  Temp: TRbNode;
begin
  while (Node <> Root) and Node.IsBlack do
  begin
    if Node = Node.Parent.Left then
    begin
      // Node is the left child of its parent,
      // and Temp is the other child.
      Temp := Node.Parent.Right;
      if Temp.IsRed then
      begin
        Temp.Color := rbBlack;
        Node.Parent.Color := rbRed;
        LeftRotate(Node.Parent);
        Temp := Node.Parent.Right;
      end;
      if Temp.Left.IsBlack and Temp.Right.IsBlack then
      begin
        Temp.Color := rbRed;
        Node := Node.Parent;
      end
      else
      begin
        if Temp.Right.IsBlack then
        begin
          Temp.Left.Color := rbBlack;
          Temp.Color := rbRed;
          RightRotate(Temp);
          Temp := Node.Parent.Right;
        end;
        Temp.Color := Node.Parent.Color;
        Node.Parent.Color := rbBlack;
        Temp.Right.Color := rbBlack;
        LeftRotate(Node.Parent);
        Node := Root;
      end;
    end
    else
    begin
      // Node is the right child of its parent,
      // and Temp is the other child.
      Temp := Node.Parent.Left;
      if Temp.IsRed then
      begin
        Temp.Color := rbBlack;
        Node.Parent.Color := rbRed;
        RightRotate(Node.Parent);
        Temp := Node.Parent.Left;
      end;
      if Temp.Right.IsBlack and Temp.Left.IsBlack then
      begin
        Temp.Color := rbRed;
        Node := Node.Parent;
      end
      else
      begin
        if Temp.Left.IsBlack then
        begin
          Temp.Right.Color := rbBlack;
          Temp.Color := rbRed;
          LeftRotate(Temp);
          Temp := Node.Parent.Left;
        end;
        Temp.Color := Node.Parent.Color;
        Node.Parent.Color := rbBlack;
        Temp.Left.Color := rbBlack;
        RightRotate(Node.Parent);
        Node := Root;
      end;
    end;
  end;
  Node.Color := rbBlack;
end;

// In CLR, x=Node, y=Neighbor.
// Rotate the subtree so Node.Right becomes the new parent of Node.
procedure TRbTree.LeftRotate(Node: TRbNode);
var
  Neighbor: TRbNode;
begin
  Neighbor := Node.Right;

  // Neighbor's left subtree becomes Node's right subtree.
  Node.Right := Neighbor.Left;
  if Neighbor.Left <> Leaf then
    Neighbor.Left.Parent := Node;

  // Change Neighbor's parent to Node's parent. If Node was the root,
  // use Neighbor as the new root. Otherwise, fixup the parent's
  // child pointer to point to Neighbor.
  Neighbor.Parent := Node.Parent;
  if Node.Parent = Leaf then
    // Node was the root.
    Root := Neighbor
  else if Node = Node.Parent.Left then
    // Node was the parent's left child.
    Node.Parent.Left := Neighbor
  else
    // Node was the parent's right child.
    Node.Parent.Right := Neighbor;

  // Move Node to the left child of Neighbor.
  Neighbor.Left := Node;
  Node.Parent := Neighbor;
end;

// In CLR, x=Neighbor, y=Node.
// Rotate the subtree so Node.Left becomes the new parent of Node.
procedure TRbTree.RightRotate(Node: TRbNode);
var
  Neighbor: TRbNode;
begin
  Neighbor := Node.Left;

  // Neighbor's right subtree becomes Node's left subtree.
  Node.Left := Neighbor.Right;
  if Neighbor.Right <> Leaf then
    Neighbor.Right.Parent := Node;

  // Change Neighbor's parent to Node's parent. If Node was the root,
  // use Neighbor as the new root. Otherwise, fixup the parent's
  // child pointer to point to Neighbor.
  Neighbor.Parent := Node.Parent;
  if Node.Parent = Leaf then
    // Node was the root.
    Root := Neighbor
  else if Node = Node.Parent.Left then
    // Node was the parent's left child.
    Node.Parent.Left := Neighbor
  else
    // Node was the parent's right child.
    Node.Parent.Right := Neighbor;

  // Move Node to the right child of Neighbor.
  Neighbor.Right := Node;
  Node.Parent := Neighbor;
end;

procedure TRbTree.Clear;

  procedure ClearTree(Node: TRbNode);
  begin
    if Node <> Leaf then
    begin
      ClearTree(Node.Left);
      ClearTree(Node.Right);
      Dec(fCount);
      Node.Free;
    end;
  end;

begin
  ClearTree(Root);
  Root := Leaf;
  Assert(Count = 0);
end;

function TRbTree.GetDuplicates: TDuplicates;
begin
  Result := dupIgnore;
end;

function TRbTree.IsSorted: Boolean;
begin
  Result := True
end;

// Return the predecessor of a node or Leaf if Node is
// the first node in the tree.
function TRbTree.Pred(Node: TRbNode): TRbNode;
begin
  if Node.Left <> Leaf then
  begin
    // Descend the right-children to the end.
    Result := Node.Left;
    while Result.Right <> Leaf do
      Result := Result.Right;
  end
  else
  begin
    // Backtrack up the tree, as long as the enumerator has descended
    // the left branch. Stop when it descended the right branch. The next
    // node is that parent node.
    while (Node <> Root) and (Node.Parent.Left = Node) do
      Node := Node.Parent;
    if Node = Root then
      Result := Leaf
    else
      Result := Node.Parent;
  end;
end;

// Return the succecessor of a node or Leaf if Node is
// the last node in the tree.
function TRbTree.Succ(Node: TRbNode): TRbNode;
begin
  if Node.Right <> Leaf then
  begin
    // Descend the left-children to the end.
    Result := Node.Right;
    while Result.Left <> Leaf do
      Result := Result.Left;
  end
  else
  begin
    // Backtrack up the tree, as long as the enumerator has descended
    // the right branch. Stop when it descended the left branch. The next
    // node is that parent node.
    while (Node <> Root) and (Node.Parent.Right = Node) do
      Node := Node.Parent;
    if Node = Root then
      Result := Leaf
    else
      Result := Node.Parent;
  end;
end;


{ TEnumRbTree }

constructor TEnumRbTree.Create(Collection: TCollection);
begin
  if not (Collection is TRbTree) then
    raise EInvalidCast.CreateFmt(sInvalidCollection, [Collection.ClassName]);
  inherited;
  Node := RbTree.Leaf;
end;

procedure TEnumRbTree.Clone(out Enum: IEnumerator);
var
  EnumObject: TEnumRbTree;
begin
  EnumObject := TEnumRbTree.Create(CollectionObject);
  EnumObject.Node := Node;
  Enum := EnumObject;
end;

function TEnumRbTree.GetRbTree: TRbTree;
begin
  Result := TRbTree(CollectionObject);
end;

function TEnumRbTree.Next(out Item: ICollectible): Boolean;
begin
  if Node = nil then
  begin
    // A nil value in Node means the enumerator has finished enumerating
    // all the nodes in the tree.
    Result := False;
    Exit;
  end;

  if Node = RbTree.Leaf then
  begin
    // Find the first node in the tree.
    Node := RbTree.Root;
    while (Node <> RbTree.Leaf) and (Node.Left <> RbTree.Leaf) do
      Node := Node.Left;
  end
  else
    Node := RbTree.Succ(Node);

  Result := Node <> RbTree.Leaf;
  if Result then
    Item := Node.Item
  else
    Node := nil; // remember that the enumerator has finished
end;

// Reset the enumerator to the beginning of the collection.
procedure TEnumRbTree.Reset;
begin
  Node := RbTree.Leaf;
end;

procedure TEnumRbTree.Remove;
var
  NodeToDelete: TRbNode;
begin
  // Make sure the enumerator is initialized.
  if (Node = nil) or (Node = RbTree.Leaf) then
    CantRemoveError;

  NodeToDelete := Node;

  // Reposition the enumerator, so Next will pick up the node
  // that is the successor to the node about to be deleted.
  // That means setting Node to its predecessor.
  Node := RbTree.Pred(Node);

  RbTree.RemoveNode(NodeToDelete);
end;

// Return True if another node is available to be returned by Next.
// If Node is nil, the enumerator has finished. If it is Leaf, the
// enumerator has not yet begun, so it checks whether the tree is empty.
function TEnumRbTree.HasNext: Boolean;
begin
  if Node = nil then
    Result := False
  else if Node <> RbTree.Leaf then
    Result := RbTree.Succ(Node) <> RbTree.Leaf
  else
    Result := not RbTree.IsEmpty
end;


{ TRbLeaf }

constructor TRbLeaf.Create;
begin
  inherited Create(nil, Self, Self, rbBlack);
end;

{$ifndef DIRECT}
function TRbLeaf.GetColor: TRbColor;
begin
  Result := rbBlack
end;
{$endif}

procedure TRbLeaf.SetColor(Color: TRbColor);
begin
  // do nothing: the color is always black.
end;



{ TRbTreeMap }

procedure TRbTreeMap.CreateCollection(out Collection: ICollection);
var
  TreeSet: TRbTreeSet;
begin
  TreeSet := TRbTreeSet.Create;
  Collection := TreeSet;
  fTree := TreeSet.Tree;
end;

procedure TRbTreeMap.CreateKeysCollection(out Keys: ISet);
begin
  Keys := TRbTreeSet.Create;
end;

procedure TRbTreeMap.CreateValuesCollection(out Values: ICollection);
begin
  Values := TArrayList.Create;
end;

{ TRbTreeSet }

procedure TRbTreeSet.CreateCollection(out Collection: ICollection);
begin
  fTree := TRbTree.Create;
  Collection := fTree;
end;

end.
